# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import logging
import pandas as pd
from tigeropen.common.consts import Market, QuoteRight, FinancialReportPeriodType, Valuation, \
    Income, Balance, CashFlow, BalanceSheetRatio, Growth, Leverage, Profitability, IndustryLevel

from tigeropen.quote.quote_client import QuoteClient

from tigeropen.examples.client_config import get_client_config

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s %(message)s',
                    filemode='a', )
logger = logging.getLogger('TigerOpenApi')

client_config = get_client_config()
openapi_client = QuoteClient(client_config, logger=logger)


def get_quote():
    # 抢占行情权限
    quote_permissions = openapi_client.grab_quote_permission()
    print(quote_permissions)

    market_status_list = openapi_client.get_market_status(Market.US)
    print(market_status_list)
    briefs = openapi_client.get_briefs(symbols=['AAPL', '00700', '600519'], include_ask_bid=True, right=QuoteRight.BR)
    print(briefs)
    briefs = openapi_client.get_stock_briefs(symbols=['AAPL', '00700', '600519'])
    print(briefs)
    metas = openapi_client.get_trade_metas(symbols=['00700'])
    print(metas)
    timelines = openapi_client.get_timeline(['AAPL'], include_hour_trading=True)
    print(timelines)
    bars = openapi_client.get_bars(['AAPL'])
    print(bars)
    ticks = openapi_client.get_trade_ticks(['00700'])
    print(ticks)
    short_interest = openapi_client.get_short_interest(['GOOG', 'AAPL', '00700'])
    print(short_interest)
    # 获取深度行情
    order_book = openapi_client.get_depth_quote(['02828'], Market.HK)
    print(order_book)
    # 股票详情
    stock_details = openapi_client.get_stock_details(['AAPL', '03690'])
    print(stock_details)


def get_option_quote():
    symbol = 'AAPL'
    expirations = openapi_client.get_option_expirations(symbols=[symbol])
    if len(expirations) > 1:
        expiry = int(expirations[expirations['symbol'] == symbol].at[0, 'timestamp'])
        chains = openapi_client.get_option_chain(symbol, expiry)
        print(chains)

    briefs = openapi_client.get_option_briefs(['AAPL  190104C00121000'])
    print(briefs)
    bars = openapi_client.get_option_bars(['AAPL  190104P00134000'])
    print(bars)
    ticks = openapi_client.get_option_trade_ticks(['AAPL  190104P00134000'])
    print(ticks)


def get_future_quote():
    exchanges = openapi_client.get_future_exchanges()
    print(exchanges)
    bars = openapi_client.get_future_bars(['CN1901'], begin_time=-1, end_time=1545105097358)
    print(bars)
    ticks = openapi_client.get_future_trade_ticks(['CN1901'])
    print(ticks)
    contracts = openapi_client.get_future_contracts('CME')
    print(contracts)
    trading_times = openapi_client.get_future_trading_times('CN1901', trading_date=1545049282852)
    print(trading_times)
    briefs = openapi_client.get_future_brief(['ES1906', 'CN1901'])
    print(briefs)


def get_fundamental():
    """获取基础数据"""
    
    # 日级财务数据
    financial_daily = openapi_client.get_financial_daily(symbols=['AAPL', 'MSFT'],
                                                         market=Market.US,
                                                         fields=[Valuation.shares_outstanding],
                                                         begin_date='2019-01-01',
                                                         end_date='2019-01-10')
    print(financial_daily)
    
    # 财报数据(季报或年报)
    financial_report = openapi_client.get_financial_report(symbols=['AAPL', 'GOOG'],
                                                           market=Market.US,
                                                           fields=[Income.revenues, CashFlow.cash_from_investing],
                                                           period_type=FinancialReportPeriodType.ANNUAL)
    print(financial_report)

    # 拆合股数据
    corporate_split = openapi_client.get_corporate_split(symbols=['UVXY', 'TQQQ'],
                                                         market=Market.US,
                                                         begin_date='2017-01-01',
                                                         end_date='2019-01-01')
    print(corporate_split)

    # 派息数据
    corporate_dividend = openapi_client.get_corporate_dividend(symbols=['MSFT', 'AAPL'],
                                                               market=Market.US,
                                                               begin_date='2018-01-01',
                                                               end_date='2019-01-01')
    print(corporate_dividend)

    # 财报日历
    earnings_calendar = openapi_client.get_corporate_earnings_calendar(Market.US, '2020-01-01', '2020-02-01')
    print(earnings_calendar)

    # 行业数据
    # 获取行业列表
    industries = openapi_client.get_industry_list(IndustryLevel.GGROUP)
    print(industries)
    # 获取某行业下公司列表
    industry_stocks = openapi_client.get_industry_stocks(50101020)
    print(industry_stocks)
    # 获取某股票的行业
    stock_industry = openapi_client.get_stock_industry('AAPL', Market.US)
    print(stock_industry)


if __name__ == '__main__':
    with pd.option_context('display.max_rows', None, 'display.max_columns', None):
        get_quote()
        get_option_quote()
        get_future_quote()
        get_fundamental()



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import pandas as pd

from tigeropen.common.util.string_utils import get_string
from tigeropen.common.response import TigerResponse


class FutureExchangeResponse(TigerResponse):
    def __init__(self):
        super(FutureExchangeResponse, self).__init__()
        self.exchanges = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(FutureExchangeResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            items = []
            for item in self.data:
                code, name, zone = None, None, None
                for key, value in item.items():
                    if value is None:
                        continue
                    if key == 'code':
                        code = get_string(value)
                    elif key == 'name':
                        name = get_string(value)
                    elif key == 'zoneId':
                        zone = get_string(value)
                items.append([code, name, zone])
            self.exchanges = pd.DataFrame(items, columns=['code', 'name', 'zone'])



# -*- coding: utf-8 -*-


class FinancialDailyParams(object):
    def __init__(self):
        self._symbols = None
        self._market = None
        self._period_type = None
        self._fields = None
        self._begin_date = None
        self._end_date = None

    @property
    def symbols(self):
        return self._symbols

    @symbols.setter
    def symbols(self, value):
        self._symbols = value

    @property
    def market(self):
        return self._market

    @market.setter
    def market(self, value):
        self._market = value

    @property
    def period_type(self):
        return self._period_type

    @period_type.setter
    def period_type(self, value):
        self._period_type = value

    @property
    def fields(self):
        return self._fields

    @fields.setter
    def fields(self, value):
        self._fields = value

    @property
    def begin_date(self):
        return self._begin_date

    @begin_date.setter
    def begin_date(self, value):
        self._begin_date = value

    @property
    def end_date(self):
        return self._end_date

    @end_date.setter
    def end_date(self, value):
        self._end_date = value

    def to_openapi_dict(self):
        params = dict()

        if self.symbols:
            params['symbols'] = self.symbols

        if self.market:
            params['market'] = self.market

        if self.period_type:
            params['period_type'] = self.period_type

        if self.fields:
            params['fields'] = self.fields

        if self.begin_date:
            params['begin_date'] = self.begin_date

        if self.end_date:
            params['end_date'] = self.end_date

        return params


class FinancialReportParams(object):
    def __init__(self):
        self._symbols = None
        self._market = None
        self._fields = None
        self._period_type = None

    @property
    def symbols(self):
        return self._symbols

    @symbols.setter
    def symbols(self, value):
        self._symbols = value

    @property
    def market(self):
        return self._market

    @market.setter
    def market(self, value):
        self._market = value

    @property
    def fields(self):
        return self._fields

    @fields.setter
    def fields(self, value):
        self._fields = value

    @property
    def period_type(self):
        return self._period_type

    @period_type.setter
    def period_type(self, value):
        self._period_type = value

    def to_openapi_dict(self):
        params = dict()

        if self.symbols:
            params['symbols'] = self.symbols

        if self.market:
            params['market'] = self.market

        if self.fields:
            params['fields'] = self.fields

        if self.period_type:
            params['period_type'] = self.period_type

        return params


class CorporateActionParams(object):
    def __init__(self):
        self._symbols = None
        self._market = None
        self._action_type = None
        self._begin_date = None
        self._end_date = None

    @property
    def symbols(self):
        return self._symbols

    @symbols.setter
    def symbols(self, value):
        self._symbols = value

    @property
    def market(self):
        return self._market

    @market.setter
    def market(self, value):
        self._market = value

    @property
    def action_type(self):
        return self._action_type

    @action_type.setter
    def action_type(self, value):
        self._action_type = value

    @property
    def begin_date(self):
        return self._begin_date

    @begin_date.setter
    def begin_date(self, value):
        self._begin_date = value

    @property
    def end_date(self):
        return self._end_date

    @end_date.setter
    def end_date(self, value):
        self._end_date = value

    def to_openapi_dict(self):
        params = dict()

        if self.symbols:
            params['symbols'] = self.symbols

        if self.market:
            params['market'] = self.market

        if self.action_type:
            params['action_type'] = self.action_type

        if self.begin_date:
            params['begin_date'] = self.begin_date

        if self.end_date:
            params['end_date'] = self.end_date

        return params


class IndustryParams(object):
    def __init__(self):
        self._industry_level = None
        self._industry_id = None
        self._market = None
        self._symbol = None

    @property
    def industry_level(self):
        return self._industry_level

    @industry_level.setter
    def industry_level(self, value):
        self._industry_level = value

    @property
    def industry_id(self):
        return self._industry_id

    @industry_id.setter
    def industry_id(self, value):
        self._industry_id = value

    @property
    def market(self):
        return self._market

    @market.setter
    def market(self, value):
        self._market = value

    @property
    def symbol(self):
        return self._symbol

    @symbol.setter
    def symbol(self, value):
        self._symbol = value

    def to_openapi_dict(self):
        params = dict()

        if self.industry_level:
            params['industry_level'] = self.industry_level

        if self.industry_id:
            params['industry_id'] = self.industry_id

        if self.market:
            params['market'] = self.market

        if self.symbol:
            params['symbol'] = self.symbol

        return params



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""

import base64
import binascii
import json
import logging
import sys

import rsa

from tigeropen.common.consts import PYTHON_VERSION_3
from tigeropen.common.util.string_utils import add_start_end


def get_sign_content(all_params):
    sign_content = ""
    for (k, v) in sorted(all_params.items()):
        value = v
        if not isinstance(value, str):
            value = json.dumps(value, ensure_ascii=False)
        sign_content += ("&" + k + "=" + value)
    sign_content = sign_content[1:]
    return sign_content


def read_private_key(key_file):
    """
    Pem key
    :param key_file:
    :return:
    """
    key_str = open(key_file, 'r').read()
    return key_str.replace('-----BEGIN RSA PRIVATE KEY-----\n', '').replace(
        '\n-----END RSA PRIVATE KEY-----', '').strip()


def read_public_key(key_file):
    """
    Pem key
    :param key_file:
    :return:
    """
    key_str = open(key_file, 'r').read()
    return key_str.replace('-----BEGIN PUBLIC KEY-----\n', '').replace('\n-----END PUBLIC KEY-----', '').strip()


def fill_private_key_marker(private_key):
    return add_start_end(private_key, "-----BEGIN RSA PRIVATE KEY-----\n", "\n-----END RSA PRIVATE KEY-----")


def fill_public_key_marker(public_key):
    return add_start_end(public_key, "-----BEGIN PUBLIC KEY-----\n", "\n-----END PUBLIC KEY-----")


def sign_with_rsa(private_key, sign_content, charset):
    if PYTHON_VERSION_3:
        sign_content = sign_content.encode(charset)
    try:
        private_key = rsa.PrivateKey.load_pkcs1(fill_private_key_marker(private_key), format='PEM')
    except binascii.Error:
        logging.error("私钥格式错误, 请参考文档进行修改. https://quant.itiger.com/openapi/py-docs/zh-cn/docs/intro/quickstart.html ")
        sys.exit(1)

    signature = rsa.sign(sign_content, private_key, 'SHA-1')

    sign = base64.b64encode(signature)
    if PYTHON_VERSION_3:
        sign = str(sign, encoding=charset)
    return sign


def verify_with_rsa(public_key, message, sign):
    public_key = fill_public_key_marker(public_key)
    sign = base64.b64decode(sign)
    return rsa.verify(message, sign, rsa.PublicKey.load_pkcs1_openssl_pem(public_key))



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
import pytz

eastern = pytz.timezone('US/Eastern')
china = pytz.timezone('Asia/Shanghai')
hongkong = pytz.timezone('Asia/Hong_Kong')


def has_value(m, key):
    if not m:
        return False
    if not (key in m):
        return False
    if not m[key]:
        return False
    return True



# -*- coding: utf-8 -*-
"""
Created on 2018/10/30

@author: gaoan
"""
import sys
import json
from collections import defaultdict

import stomp
import six
import logging
from stomp.exception import ConnectFailedException

from tigeropen.common.consts.push_destinations import QUOTE, QUOTE_DEPTH, QUOTE_FUTURE, QUOTE_OPTION, TRADE_ASSET, \
    TRADE_ORDER, TRADE_POSITION
from tigeropen.common.util.signature_utils import sign_with_rsa
from tigeropen.common.util.order_utils import get_order_status
from tigeropen.common.consts.push_types import RequestType, ResponseType
from tigeropen.common.consts.quote_keys import QuoteChangeKey, QuoteKeyType
from tigeropen.common.consts import OrderStatus

HOUR_TRADING_QUOTE_KEYS_MAPPINGS = {'hourTradingLatestPrice': 'latest_price', 'hourTradingPreClose': 'pre_close',
                                    'hourTradingLatestTime': 'latest_time', 'hourTradingVolume': 'volume',
                                    }
QUOTE_KEYS_MAPPINGS = {field.value: field.name for field in QuoteChangeKey}  # like {'askPrice': 'ask_price'}
QUOTE_KEYS_MAPPINGS.update(HOUR_TRADING_QUOTE_KEYS_MAPPINGS)
PRICE_FIELDS = {'open', 'high', 'low', 'close', 'prev_close', 'ask_price', 'bid_price', 'latest_price'}

ASSET_KEYS_MAPPINGS = {'buyingPower': 'buying_power', 'cashBalance': 'cash',
                       'grossPositionValue': 'gross_position_value',
                       'netLiquidation': 'net_liquidation', 'equityWithLoan': 'equity_with_loan',
                       'initMarginReq': 'initial_margin_requirement',
                       'maintMarginReq': 'maintenance_margin_requirement',
                       'availableFunds': 'available_funds', 'excessLiquidity': 'excess_liquidity',
                       'dayTradesRemaining': 'day_trades_remaining', 'currency': 'currency', 'segment': 'segment'}

POSITION_KEYS_MAPPINGS = {'averageCost': 'average_cost', 'position': 'quantity', 'latestPrice': 'market_price',
                          'marketValue': 'market_value', 'orderType': 'order_type', 'realizedPnl': 'realized_pnl',
                          'unrealizedPnl': 'unrealized_pnl', 'secType': 'sec_type', 'localSymbol': 'local_symbol',
                          'originSymbol': 'origin_symbol', 'contractId': 'contract_id', 'symbol': 'symbol',
                          'currency': 'currency', 'strike': 'strike', 'expiry': 'expiry', 'right': 'right',
                          'segment': 'segment', 'identifier': 'identifier'}

ORDER_KEYS_MAPPINGS = {'parentId': 'parent_id', 'orderId': 'order_id', 'orderType': 'order_type',
                       'limitPrice': 'limit_price', 'auxPrice': 'aux_price', 'avgFillPrice': 'avg_fill_price',
                       'totalQuantity': 'quantity', 'filledQuantity': 'filled', 'lastFillPrice': 'last_fill_price',
                       'realizedPnl': 'realized_pnl', 'secType': 'sec_type', 'symbol': 'symbol',
                       'remark': 'reason', 'localSymbol': 'local_symbol', 'originSymbol': 'origin_symbol',
                       'outsideRth': 'outside_rth', 'timeInForce': 'time_in_force', 'openTime': 'order_time',
                       'latestTime': 'trade_time', 'contractId': 'contract_id', 'trailStopPrice': 'trail_stop_price',
                       'trailingPercent': 'trailing_percent', 'percentOffset': 'percent_offset', 'action': 'action',
                       'status': 'status', 'currency': 'currency', 'remaining': 'remaining', 'id': 'id',
                       'segment': 'segment', 'identifier': 'identifier', 'replaceStatus': 'replace_status'}

if sys.platform == 'linux' or sys.platform == 'linux2':
    KEEPALIVE = True
else:
    KEEPALIVE = False


class PushClient(stomp.ConnectionListener):
    def __init__(self, host, port, use_ssl=True, connection_timeout=120, auto_reconnect=True,
                 heartbeats=(30 * 1000, 30 * 1000)):
        """
        :param host:
        :param port:
        :param use_ssl:
        :param connection_timeout: second
        :param auto_reconnect:
        :param heartbeats: tuple of millisecond
        """
        self.host = host
        self.port = port
        self.use_ssl = use_ssl
        self._tiger_id = None
        self._private_key = None
        self._sign = None
        self._stomp_connection = None
        self._destination_counter_map = defaultdict(lambda: 0)

        self.subscribed_symbols = None
        self.quote_changed = None
        self.asset_changed = None
        self.position_changed = None
        self.order_changed = None
        self.connect_callback = None
        self.disconnect_callback = None
        self.subscribe_callback = None
        self.unsubscribe_callback = None
        self.error_callback = None
        self._connection_timeout = connection_timeout
        self._auto_reconnect = auto_reconnect
        self._heartbeats = heartbeats

    def _connect(self):
        try:
            if self._stomp_connection:
                self._stomp_connection.remove_listener('push')
                self._stomp_connection.transport.cleanup()
        except:
            pass

        self._stomp_connection = stomp.Connection12(host_and_ports=[(self.host, self.port), ], use_ssl=self.use_ssl,
                                                    keepalive=KEEPALIVE, timeout=self._connection_timeout,
                                                    heartbeats=self._heartbeats)
        self._stomp_connection.set_listener('push', self)
        self._stomp_connection.start()
        try:
            self._stomp_connection.connect(self._tiger_id, self._sign, wait=True)
        except ConnectFailedException as e:
            raise e

    def connect(self, tiger_id, private_key):
        self._tiger_id = tiger_id
        self._private_key = private_key
        self._sign = sign_with_rsa(self._private_key, self._tiger_id, 'utf-8')
        self._connect()

    def disconnect(self):
        if self._stomp_connection:
            self._stomp_connection.disconnect()

    def on_connected(self, headers, body):
        if self.connect_callback:
            self.connect_callback()

    def on_disconnected(self):
        if self.disconnect_callback:
            self.disconnect_callback()
        elif self._auto_reconnect:
            self._connect()

    def on_message(self, headers, body):
        """
        Called by the STOMP connection when a MESSAGE frame is received.

        :param dict headers: a dictionary containing all headers sent by the server as key/value pairs.
        :param body: the frame's payload - the message body.
        """
        try:
            response_type = headers.get('ret-type')
            if response_type == str(ResponseType.GET_SUB_SYMBOLS_END.value):
                if self.subscribed_symbols:
                    data = json.loads(body)
                    limit = data.get('limit')
                    symbols = data.get('subscribedSymbols')
                    used = data.get('used')
                    symbol_focus_keys = data.get('symbolFocusKeys')
                    focus_keys = dict()
                    for sym, keys in symbol_focus_keys.items():
                        keys = set(QUOTE_KEYS_MAPPINGS.get(key, key) for key in keys)
                        focus_keys[sym] = list(keys)
                    self.subscribed_symbols(symbols, focus_keys, limit, used)
            elif response_type == str(ResponseType.GET_QUOTE_CHANGE_END.value):
                if self.quote_changed:
                    data = json.loads(body)
                    hour_trading = False
                    if 'hourTradingLatestPrice' in data:
                        hour_trading = True
                    if 'symbol' in data:
                        symbol = data.get('symbol')
                        offset = data.get('offset', 0)
                        items = []
                        # 期货行情推送的价格都乘了 10 的 offset 次方变成了整数, 需要除回去变为正常单位的价格
                        if offset:
                            for key, value in data.items():
                                if (key == 'latestTime' or key == 'hourTradingLatestTime') and \
                                        isinstance(value, six.string_types):
                                    continue
                                if key in QUOTE_KEYS_MAPPINGS:
                                    key = QUOTE_KEYS_MAPPINGS.get(key)
                                    if key in PRICE_FIELDS:
                                        value /= 10 ** offset
                                    elif key == 'minute':
                                        minute_item = dict()
                                        for m_key, m_value in value.items():
                                            if m_key in {'p', 'h', 'l'}:
                                                m_value /= 10 ** offset
                                            minute_item[m_key] = m_value
                                            value = minute_item
                                    items.append((key, value))
                        else:
                            for key, value in data.items():
                                if (key == 'latestTime' or key == 'hourTradingLatestTime') and \
                                        isinstance(value, six.string_types):
                                    continue
                                if key in QUOTE_KEYS_MAPPINGS:
                                    key = QUOTE_KEYS_MAPPINGS.get(key)
                                    items.append((key, value))
                        if items:
                            self.quote_changed(symbol, items, hour_trading)
            elif response_type == str(ResponseType.SUBSCRIBE_ASSET.value):
                if self.asset_changed:
                    data = json.loads(body)
                    if 'account' in data:
                        account = data.get('account')
                        items = []
                        for key, value in data.items():
                            if key in ASSET_KEYS_MAPPINGS:
                                items.append((ASSET_KEYS_MAPPINGS.get(key), value))
                        if items:
                            self.asset_changed(account, items)
            elif response_type == str(ResponseType.SUBSCRIBE_POSITION.value):
                if self.position_changed:
                    data = json.loads(body)
                    if 'account' in data:
                        account = data.get('account')
                        items = []
                        for key, value in data.items():
                            if key in POSITION_KEYS_MAPPINGS:
                                items.append((POSITION_KEYS_MAPPINGS.get(key), value))
                        if items:
                            self.position_changed(account, items)
            elif response_type == str(ResponseType.SUBSCRIBE_ORDER_STATUS.value):
                if self.order_changed:
                    data = json.loads(body)
                    if 'account' in data:
                        account = data.get('account')
                        items = []
                        for key, value in data.items():
                            if key in ORDER_KEYS_MAPPINGS:
                                if key == 'status':
                                    value = get_order_status(value)
                                    # 部分成交 (服务端推送 'Submitted' 状态)
                                    if value == OrderStatus.HELD and data.get('filledQuantity'):
                                        value = OrderStatus.PARTIALLY_FILLED
                                items.append((ORDER_KEYS_MAPPINGS.get(key), value))
                        if items:
                            self.order_changed(account, items)
            elif response_type == str(ResponseType.GET_SUBSCRIBE_END.value):
                if self.subscribe_callback:
                    self.subscribe_callback(headers.get('destination'), json.loads(body))
            elif response_type == str(ResponseType.GET_CANCEL_SUBSCRIBE_END.value):
                if self.unsubscribe_callback:
                    self.unsubscribe_callback(headers.get('destination'), json.loads(body))
            elif response_type == str(ResponseType.ERROR_END.value):
                if self.error_callback:
                    self.error_callback(body)
        except Exception as e:
            logging.error(e, exc_info=True)

    def on_error(self, headers, body):
        if self.error_callback:
            self.error_callback(body)

    def _update_subscribe_id(self, destination):
        self._destination_counter_map[destination] += 1

    def _get_subscribe_id(self, destination):
        return 'sub-' + str(self._destination_counter_map[destination])

    def subscribe_asset(self, account=None):
        """
        订阅账户资产更新
        :return:
        """
        return self._handle_trade_subscribe(TRADE_ASSET, 'Asset', account)

    def unsubscribe_asset(self, id=None):
        """
        退订账户资产更新
        :return:
        """
        self._handle_trade_unsubscribe(TRADE_ASSET, 'Asset', sub_id=id)

    def subscribe_position(self, account=None):
        """
        订阅账户持仓更新
        :return:
        """
        return self._handle_trade_subscribe(TRADE_POSITION, 'Position', account)

    def unsubscribe_position(self, id=None):
        """
        退订账户持仓更新
        :return:
        """
        self._handle_trade_unsubscribe(TRADE_POSITION, 'Position', sub_id=id)

    def subscribe_order(self, account=None):
        """
        订阅账户订单更新
        :return:
        """
        return self._handle_trade_subscribe(TRADE_ORDER, 'OrderStatus', account)

    def unsubscribe_order(self, id=None):
        """
        退订账户订单更新
        :return:
        """
        self._handle_trade_unsubscribe(TRADE_ORDER, 'OrderStatus', sub_id=id)

    def subscribe_quote(self, symbols, quote_key_type=QuoteKeyType.TRADE, focus_keys=None):
        """
        订阅行情更新
        :param symbols:
        :param quote_key_type: 行情类型, 值为 common.consts.quote_keys.QuoteKeyType 枚举类型
        :param focus_keys: 行情 key
        :return:
        """
        extra_headers = dict()
        if focus_keys:
            keys = list()
            for key in focus_keys:
                if isinstance(key, six.string_types):
                    keys.append(key)
                else:
                    keys.append(key.value)
            extra_headers['keys'] = ','.join(keys)
        elif quote_key_type and quote_key_type.value:
            extra_headers['keys'] = quote_key_type.value
        return self._handle_quote_subscribe(destination=QUOTE, subscription='Quote', symbols=symbols,
                                            extra_headers=extra_headers)

    def subscribe_depth_quote(self, symbols):
        """
        订阅深度行情
        :param symbols: symbol列表
        :return:
        """
        return self._handle_quote_subscribe(destination=QUOTE_DEPTH, subscription='QuoteDepth', symbols=symbols)

    def subscribe_option(self, symbols):
        """
        订阅期权行情
        :param symbols: symbol列表
        :return:
        """
        return self._handle_quote_subscribe(destination=QUOTE_OPTION, subscription='Option', symbols=symbols)

    def subscribe_future(self, symbols):
        """
        订阅期货行情
        :param symbols: symbol列表
        :return:
        """
        return self._handle_quote_subscribe(destination=QUOTE_FUTURE, subscription='Future', symbols=symbols)

    def query_subscribed_quote(self):
        """
        查询已订阅行情的合约
        :return:
        """
        headers = dict()
        headers['destination'] = QUOTE
        headers['req-type'] = RequestType.REQ_SUB_SYMBOLS.value
        self._stomp_connection.send(QUOTE, "{}", headers=headers)

    def unsubscribe_quote(self, symbols=None, id=None):
        """
        退订行情更新
        :return:
        """
        self._handle_quote_unsubscribe(destination=QUOTE, subscription='Quote', sub_id=id, symbols=symbols)

    def unsubscribe_depth_quote(self, symbols=None, id=None):
        """
        退订深度行情更新
        :return:
        """
        self._handle_quote_unsubscribe(destination=QUOTE_DEPTH, subscription='AskBid', sub_id=id, symbols=symbols)

    def _handle_trade_subscribe(self, destination, subscription, account=None, extra_headers=None):
        if extra_headers is None:
            extra_headers = dict()
        if account is not None:
            extra_headers['account'] = account
        return self._handle_subscribe(destination=destination, subscription=subscription, extra_headers=extra_headers)

    def _handle_quote_subscribe(self, destination, subscription, symbols=None, extra_headers=None):
        if extra_headers is None:
            extra_headers = dict()
        if symbols is not None:
            extra_headers['symbols'] = ','.join(symbols)
        return self._handle_subscribe(destination=destination, subscription=subscription, extra_headers=extra_headers)

    def _handle_trade_unsubscribe(self, destination, subscription, sub_id=None):
        self._handle_unsubscribe(destination=destination, subscription=subscription, sub_id=sub_id)

    def _handle_quote_unsubscribe(self, destination, subscription, sub_id=None, symbols=None):
        extra_headers = dict()
        if symbols is not None:
            extra_headers['symbols'] = ','.join(symbols)
        self._handle_unsubscribe(destination=destination, subscription=subscription, sub_id=sub_id,
                                 extra_headers=extra_headers)

    def _handle_subscribe(self, destination, subscription, extra_headers=None):
        headers = dict()
        headers['destination'] = destination
        headers['subscription'] = subscription
        self._update_subscribe_id(destination)
        sub_id = self._get_subscribe_id(destination)
        headers['id'] = sub_id

        if extra_headers is not None:
            headers.update(extra_headers)
        self._stomp_connection.subscribe(destination, id=sub_id, headers=headers)
        return sub_id

    def _handle_unsubscribe(self, destination, subscription, sub_id=None, extra_headers=None):
        headers = dict()
        headers['destination'] = destination
        headers['subscription'] = subscription
        id_ = sub_id if sub_id is not None else self._get_subscribe_id(destination)
        headers['id'] = id_
        if extra_headers:
            headers.update(extra_headers)

        self._stomp_connection.unsubscribe(id=id_, headers=headers)



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
import threading
import platform
from enum import Enum, unique
from .quote_keys import QuoteChangeKey, QuoteKeyType
from .fundamental_fields import Valuation, Income, Balance, CashFlow, BalanceSheetRatio, Growth, \
    Leverage, Profitability

python_version = platform.python_version()

if python_version.startswith("3"):
    PYTHON_VERSION_3 = True
else:
    PYTHON_VERSION_3 = False

OPEN_API_SDK_VERSION = "2.0"

THREAD_LOCAL = threading.local()


@unique
class Market(Enum):
    """Enum for market """
    
    ALL = 'ALL'
    US = 'US'  # 美股
    HK = 'HK'  # 港股
    CN = 'CN'  # A股
    SG = 'SG'  # 新加坡


@unique
class TradingSession(Enum):
    PreMarket = 'PreMarket'  # 盘前
    Regular = 'Regular'  # 盘中
    AfterHours = 'AfterHours'  # 盘后


@unique
class SecurityType(Enum):
    """Enum for sec_type """
    
    ALL = 'ALL'
    STK = 'STK'  # 股票
    OPT = 'OPT'  # 期权
    WAR = 'WAR'  # 窝轮
    IOPT = 'IOPT'  # 权证(牛熊证)
    FUT = 'FUT'  # 期货
    FOP = 'FOP'  # 期货期权
    CASH = 'CASH'  # 外汇


@unique
class Currency(Enum):
    """Enum for currency """
    
    ALL = 'ALL'
    USD = 'USD'  # 美元
    HKD = 'HKD'  # 港币
    CNH = 'CNH'  # 离岸人民币
    SGD = 'SGD'  # 新加坡币


@unique
class Language(Enum):
    zh_CN = 'zh_CN'  # 简体中文
    zh_TW = 'zh_TW'  # 繁体中文
    en_US = 'en_US'  # 英文


@unique
class QuoteRight(Enum):
    BR = 'br'  # 前复权
    NR = 'nr'  # 不复权


@unique
class TimelinePeriod(Enum):
    DAY = 'day'  # 当天分时
    FIVE_DAYS = '5day'  # 5日分时


@unique
class BarPeriod(Enum):
    DAY = 'day'  # 日K
    WEEK = 'week'  # 周K
    MONTH = 'month'  # 月K
    YEAR = 'year'  # 年K
    ONE_MINUTE = '1min'  # 1分钟
    THREE_MINUTES = '3min'  # 3分钟
    FIVE_MINUTES = '5min'  # 5分钟
    TEN_MINUTES = '10min'  # 10分钟
    FIFTEEN_MINUTES = '15min'  # 15分钟
    HALF_HOUR = '30min'  # 30分钟
    FORTY_FIVE_MINUTES = '45min'  # 45分钟
    ONE_HOUR = '60min'  # 60分钟
    TWO_HOURS = '2hour'  # 2小时
    THREE_HOURS = '3hour'  # 3小时
    FOUR_HOURS = '4hour'  # 4小时
    SIX_HOURS = '6hour'  # 6小时


class OrderStatus(Enum):
    PENDING_NEW = 'PendingNew'
    NEW = 'Initial'  # 订单初始状态
    HELD = 'Submitted'  # 已提交
    PARTIALLY_FILLED = 'PartiallyFilled'  # 部分成交
    FILLED = 'Filled'  # 完全成交
    CANCELLED = 'Cancelled'  # 已取消
    PENDING_CANCEL = 'PendingCancel'  # 待取消
    REJECTED = 'Inactive'  # 已失效
    EXPIRED = 'Invalid'  # 非法状态


@unique
class FinancialReportPeriodType(Enum):
    """
    财报类型
    """
    ANNUAL = 'Annual'  # 年报
    QUARTERLY = 'Quarterly'  # 季报
    LTM = 'LTM'  # 最近四个季度


@unique
class CorporateActionType(Enum):
    """
    公司行动类型
    """
    SPLIT = 'split'  # 拆合股
    DIVIDEND = 'dividend'  # 分红
    EARNINGS_CALENDAR = 'earning'  # 财报日历


@unique
class IndustryLevel(Enum):
    """
    行业级别. 级别从1级到4级依次为: GSECTOR, GGROUP, GIND, GSUBIND
    """
    GSECTOR = 'GSECTOR'
    GGROUP = 'GGROUP'
    GIND = 'GIND'
    GSUBIND = 'GSUBIND'



import os
import pickle
from tigeropen.common.consts import SecurityType, Currency
from .setting import MARKET, CONTEXT_FILE_PATH

SECURITY_TYPE_MAP = {
    SecurityType.STK.name: SecurityType.STK,  # 股票
    SecurityType.OPT.name: SecurityType.OPT,  # 期权
    SecurityType.WAR.name: SecurityType.WAR,  # 窝轮
    SecurityType.IOPT.name: SecurityType.IOPT,  # 权证(牛熊证)
    SecurityType.FUT.name: SecurityType.FUT,  # 期货
    SecurityType.FOP.name: SecurityType.FOP,  # 期货期权
    SecurityType.CASH.name: SecurityType.CASH,  # 外汇
}
CURRENCY_MAP = {
    Currency.USD.name: Currency.USD,  # 美元
    Currency.HKD.name: Currency.HKD,  # 港币
    Currency.CNH.name: Currency.CNH,  # 离岸人民币
}


class Context:
    _store_exclude_list = ('account', 'asset_manager', 'position_manager', 'order_manager', 'active_order_manager',
                           'contract_map', 'security_type_map', 'currency_map', 'subscribed_symbols',
                           'schedule_function', 'portfolio')
    def __init__(self):
        self.account = None
        self.asset_manager = None
        self.position_manager = None
        self.order_manager = None
        self.active_order_manager = None
        self.contract_map = {}
        self.security_type_map = {}
        self.currency_map = {}
        self.subscribed_symbols = set()

    def subscribe(self, symbols, currency=MARKET.CURRENCY, security_type='STK', exchange=None):
        if not (isinstance(symbols, list) or isinstance(symbols, set)):
            symbols = [symbols]
        for symbol in symbols:
            self.subscribed_symbols.add(symbol)
            self.security_type_map[symbol] = SECURITY_TYPE_MAP.get(security_type.upper())
            self.currency_map[symbol] = CURRENCY_MAP.get(currency.upper())

    def load(self):
        if os.path.isfile(CONTEXT_FILE_PATH):
            with open(CONTEXT_FILE_PATH, 'rb') as f:
                try:
                    loaded_state = pickle.load(f)
                except (pickle.UnpicklingError, IndexError, EOFError):
                    raise ValueError("Bad context file: {}".format(CONTEXT_FILE_PATH))
                else:
                    for k, v in loaded_state.items():
                        setattr(self, k, v)

    def store(self):
        if CONTEXT_FILE_PATH:
            context = {}
            fields_to_store = list(set(self.__dict__.keys()) - set(self._store_exclude_list))

            for field in fields_to_store:
                context[field] = getattr(self, field)

            with open(CONTEXT_FILE_PATH, 'wb') as f:
                pickle.dump(context, f)


global_context = Context()



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""

from tigeropen.common.response import TigerResponse
from tigeropen.common.util.string_utils import get_string


class SymbolsResponse(TigerResponse):
    def __init__(self):
        super(SymbolsResponse, self).__init__()
        self.symbols = []
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(SymbolsResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            self.symbols = [get_string(symbol) for symbol in self.data if symbol]



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""

import dateutil.parser as dateparser

from tigeropen.common.util.string_utils import get_string
from tigeropen.common.util.common_utils import eastern, china, hongkong
from tigeropen.quote.domain.market_status import MarketStatus
from tigeropen.common.response import TigerResponse


class MarketStatusResponse(TigerResponse):
    def __init__(self):
        super(MarketStatusResponse, self).__init__()
        self.markets = []
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(MarketStatusResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            for item in self.data:
                market, status, open_time, trading_status = None, None, None, None
                for key, value in item.items():
                    if value is None:
                        continue
                    if key == 'market':
                        market = get_string(value)
                    elif key == 'marketStatus':
                        status = get_string(value)
                    elif key == 'openTime':
                        if value.endswith(' EDT') or value.endswith(' EST'):
                            value = value[0:len(value) - 4]
                        open_time = dateparser.parse(value)
                    elif key == 'status':
                        trading_status = get_string(value)

                if open_time and market:
                    if market == 'US':
                        open_time = eastern.localize(open_time)
                    elif market == 'HK':
                        open_time = hongkong.localize(open_time)
                    elif market == 'CN':
                        open_time = china.localize(open_time)
                market_status = MarketStatus(market, status, open_time, trading_status)
                self.markets.append(market_status)



# -*- coding: utf-8 -*-
"""
Created on 2018/11/1

@author: gaoan
"""
from tigeropen.trade.domain.order import Order, OrderLeg, AlgoParams
from tigeropen.common.consts import OrderStatus


def market_order(account, contract, action, quantity):
    """
    市价单
    :param account:
    :param contract:
    :param action: BUY/SELL
    :param quantity:
    :return:
    """
    return Order(account, contract, action, 'MKT', quantity)


def limit_order(account, contract, action, quantity, limit_price):
    """
    限价单
    :param account:
    :param contract:
    :param action: BUY/SELL
    :param quantity:
    :param limit_price: 限价的价格
    :return:
    """
    return Order(account, contract, action, 'LMT', quantity, limit_price=limit_price)


def stop_order(account, contract, action, quantity, aux_price):
    """
    止损单
    :param account:
    :param contract:
    :param action: BUY/SELL
    :param quantity:
    :param aux_price: 触发止损单的价格
    :return:
    """
    return Order(account, contract, action, 'STP', quantity, aux_price=aux_price)


def stop_limit_order(account, contract, action, quantity, limit_price, aux_price):
    """
    限价止损单
    :param account:
    :param contract:
    :param action: BUY/SELL
    :param quantity:
    :param limit_price: 发出限价单的价格
    :param aux_price: 触发止损单的价格
    :return:
    """
    return Order(account, contract, action, 'STP_LMT', quantity, limit_price=limit_price, aux_price=aux_price)


def trail_order(account, contract, action, quantity, trailing_percent=None, aux_price=None):
    """
    移动止损单
    :param account:
    :param contract:
    :param action: BUY/SELL
    :param quantity:
    :param trailing_percent: 百分比 (0~100)
    :param aux_price: 价差  aux_price 和 trailing_percent 两者互斥
    :return:
    """
    return Order(account, contract, action, 'TRAIL', quantity, trailing_percent=trailing_percent, aux_price=aux_price)


def order_leg(leg_type, price, time_in_force='DAY', outside_rth=None):
    """
    附加订单
    :param leg_type: 附加订单类型. PROFIT 止盈单类型,  LOSS 止损单类型
    :param price: 附加订单价格.
    :param time_in_force: 附加订单有效期. 'DAY'（当日有效）和'GTC'（取消前有效 Good-Til-Canceled).
    :param outside_rth: 附加订单是否允许盘前盘后交易(美股专属). True 允许, False 不允许.
    """
    return OrderLeg(leg_type=leg_type, price=price, time_in_force=time_in_force, outside_rth=outside_rth)


def limit_order_with_legs(account, contract, action, quantity, limit_price, order_legs=None):
    """
    限价单 + 附加订单(仅环球账户支持)
    :param account:
    :param contract:
    :param action: BUY/SELL
    :param quantity:
    :param limit_price: 限价单价格
    :param order_legs: 附加订单列表
    :return:
    """
    if order_legs and len(order_legs) > 2:
        raise Exception('2 order legs at most')
    return Order(account, contract, action, 'LMT', quantity, limit_price=limit_price, order_legs=order_legs)


def algo_order_params(start_time=None, end_time=None, no_take_liq=None, allow_past_end_time=None, participation_rate=None):
    """
    算法订单参数
    :param start_time: 生效开始时间(时间戳 TWAP和VWAP专用)
    :param end_time: 生效结束时间(时间戳 TWAP和VWAP专用)
    :param no_take_liq: 是否尽可能减少交易次数(VWAP订单专用)
    :param allow_past_end_time: 是否允许生效时间结束后继续完成成交(TWAP和VWAP专用)
    :param participation_rate: 参与率(VWAP专用,0.01-0.5)
    :return:
    """
    return AlgoParams(start_time=start_time, end_time=end_time, no_take_liq=no_take_liq,
                      allow_past_end_time=allow_past_end_time, participation_rate=participation_rate)


def algo_order(account, contract, action, quantity, strategy, algo_params=None, limit_price=None):
    """
    算法订单
    :param account:
    :param contract:
    :param action:
    :param quantity:
    :param strategy: 交易量加权平均价格（VWAP）/时间加权平均价格(TWAP)
    :param algo_params: tigeropen.trade.domain.order.AlgoParams
    :param limit_price:
    :return:
    """
    return Order(account, contract, action, order_type=strategy, quantity=quantity, algo_params=algo_params,
                 limit_price=limit_price, outside_rth=False)


def get_order_status(value):
    """
    Invalid(-2), Initial(-1), PendingCancel(3), Cancelled(4), Submitted(5), Filled(6), Inactive(7), PendingSubmit(8)
    :param value:
    :return:
    """
    if value == -1 or value == 'Initial':
        return OrderStatus.NEW
    elif value == 2 or value == 5 or value == 8 or value == 'Submitted' or value == 'PendingSubmit':
        return OrderStatus.HELD
    elif value == 3 or value == 'PendingCancel':
        return OrderStatus.PENDING_CANCEL
    elif value == 4 or value == 'Cancelled':
        return OrderStatus.CANCELLED
    elif value == 6 or value == 'Filled':
        return OrderStatus.FILLED
    elif value == 7 or value == 'Inactive':
        return OrderStatus.REJECTED
    elif value == -2 or value == 'Invalid':
        return OrderStatus.EXPIRED

    return OrderStatus.PENDING_NEW



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
from six import text_type
from tigeropen.common.consts import OrderStatus

ORDER_FIELDS_TO_IGNORE = {'type', '_status', 'contract', '_remaining'}
ALGO_PARAMS_TAG_MAP = {'noTakeLiq': 'no_take_liq', 'startTime': 'start_time', 'endTime': 'end_time',
                       'participationRate': 'participation_rate', 'allowPastEndTime': 'allow_past_end_time'}


class Order(object):
    __slots__ = ["account", "id", "order_id", "parent_id", "order_time", "reason", "trade_time", "contract", "action",
                 "quantity", "filled", "_remaining", "avg_fill_price", "commission", "realized_pnl", "_status",
                 "trail_stop_price", "limit_price", "aux_price", "trailing_percent", "percent_offset", "action",
                 "order_type", "time_in_force", "outside_rth", "order_legs", "algo_params", "secret_key"]

    def __init__(self, account, contract, action, order_type, quantity, limit_price=None, aux_price=None,
                 trail_stop_price=None, trailing_percent=None, percent_offset=None, time_in_force=None,
                 outside_rth=None, filled=0, avg_fill_price=0, commission=None, realized_pnl=None,
                 id=None, order_id=None, parent_id=None, order_time=None, trade_time=None, order_legs=None,
                 algo_params=None, secret_key=None):
        """
        - account: 订单所属的账户
        - id: 全局订单 id
        - order_id: 账户自增订单号
        - parent_id: 主订单id, 目前只用于 TigerTrade App端的附加订单中
        - order_time: 下单时间
        - reason: 下单失败时, 会返回失败原因的描述
        - trade_time: 最新成交时间
        - action: 交易方向, 'BUY' / 'SELL'
        - quantity: 下单数量
        - filled: 成交数量
        - avg_fill_price: 包含佣金的平均成交价
        - commission: 包含佣金, 印花税, 证监会费等系列费用
        - realized_pnl: 实现盈亏
        - trail_stop_price: 跟踪止损单--触发止损单的价格
        - limit_price: 限价单价格
        - aux_price: 在止损单中, 表示出发止损单的价格, 在移动止损单中, 表示跟踪的价差
        - trailing_percent:  跟踪止损单-百分比, 取值范围为0-100
        - percent_offset: None,
        - order_type: 订单类型, 'MKT' 市价单 / 'LMT' 限价单 / 'STP' 止损单 / 'STP_LMT' 止损限价单 / 'TRAIL' 跟踪止损单
        - time_in_force: 有效期,'DAY' 日内有效 / 'GTC' 撤销前有效
        - outside_rth: 是否允许盘前盘后交易(outside of regular trading hours 美股专属). True 允许, False 不允许
        - contract: 合约对象
        - status: Order_Status 的枚举, 表示订单状态
        - remaining: 未成交的数量
        - order_legs: 附加订单列表
        - algo_params: 算法订单参数
        - secret_key: 机构交易员专有密钥
        """

        self.id = id
        self.order_id = order_id
        self.parent_id = parent_id
        self.account = account
        self.reason = None
        self.contract = contract
        self.action = action
        self.quantity = quantity
        self.filled = filled
        self._remaining = None
        self.avg_fill_price = avg_fill_price
        self.realized_pnl = realized_pnl
        self.commission = commission
        self._status = OrderStatus.NEW
        self.time_in_force = time_in_force
        self.outside_rth = outside_rth
        self.order_type = order_type
        self.limit_price = limit_price
        self.aux_price = aux_price
        self.trail_stop_price = trail_stop_price
        self.trailing_percent = trailing_percent
        self.percent_offset = percent_offset
        self.order_time = order_time
        self.trade_time = trade_time
        self.order_legs = order_legs
        self.algo_params = algo_params
        self.secret_key = secret_key

    def to_dict(self):
        dct = {name: getattr(self, name) for name in self.__slots__ if name not in ORDER_FIELDS_TO_IGNORE}

        dct['contract'] = self.contract
        if self.status:
            dct['status'] = self.status.name
        dct['remaining'] = self.remaining

        return dct

    @property
    def status(self):
        if not self.remaining and self.filled:
            return OrderStatus.FILLED
        elif self._status == OrderStatus.HELD and self.filled:
            return OrderStatus.PARTIALLY_FILLED
        else:
            return self._status

    @status.setter
    def status(self, status):
        self._status = status

    @property
    def active(self):
        return self.status in [OrderStatus.PENDING_NEW, OrderStatus.NEW, OrderStatus.PARTIALLY_FILLED,
                               OrderStatus.HELD, OrderStatus.PENDING_CANCEL]

    @property
    def remaining(self):
        return self.quantity - self.filled

    def __repr__(self):
        """
        String representation for this object.
        """
        return "Order(%s)" % self.to_dict().__repr__()

    def __unicode__(self):
        """
        Unicode representation for this object.
        """
        return text_type(repr(self))


class OrderLeg(object):
    """
    附加订单
    """

    def __init__(self, leg_type, price, time_in_force='DAY', outside_rth=None):
        """
        :param leg_type: 附加订单类型(仅限价单支持). PROFIT 止盈单类型, LOSS 止损单类型
        :param price: 附加订单价格
        :param time_in_force: 附加订单有效期. 'DAY'（当日有效）和'GTC'（取消前有效).
        :param outside_rth: 附加订单是否允许盘前盘后交易(美股专属). True 允许, False 不允许.
        """
        self.leg_type = leg_type
        self.price = price
        self.time_in_force = time_in_force
        self.outside_rth = outside_rth

    def to_dict(self):
        return self.__dict__

    def __repr__(self):
        return "OrderLeg(%s)" % self.to_dict()


class AlgoParams(object):
    """
    算法订单参数
    """
    def __init__(self, start_time=None, end_time=None, no_take_liq=None, allow_past_end_time=None,
                 participation_rate=None):
        """
        :param start_time: 生效开始时间(时间戳 TWAP和VWAP专用)
        :param end_time: 生效结束时间(时间戳 TWAP和VWAP专用)
        :param no_take_liq: 是否尽可能减少交易次数(VWAP订单专用)
        :param allow_past_end_time: 是否允许生效时间结束后继续完成成交(TWAP和VWAP专用)
        :param participation_rate: 参与率(VWAP专用,0.01-0.5)
        """
        self.start_time = start_time
        self.end_time = end_time
        self.no_take_liq = no_take_liq
        self.allow_past_end_time = allow_past_end_time
        self.participation_rate = participation_rate

    def to_dict(self):
        return self.__dict__

    @staticmethod
    def from_tags(tag_values):
        """
        :param tag_values:
        :return: AlgoParams 对象
        """
        algo_params = AlgoParams()
        if tag_values:
            for item in tag_values:
                tag = item.get('tag')
                value = item.get('value')
                setattr(algo_params, ALGO_PARAMS_TAG_MAP.get(tag), value)
            return algo_params
        return None

    def __repr__(self):
        return "AlgoParams(%s)" % self.to_dict()


# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""


class TigerResponse(object):
    def __init__(self):
        self.code = None
        self.message = None
        self.data = None

    def is_success(self):
        return self.code == 0

    def parse_response_content(self, response):
        if 'code' in response:
            self.code = response['code']
        if 'message' in response:
            self.message = response['message']
        if 'data' in response:
            self.data = response['data']
        return response



# -*- coding: utf-8 -*-
from enum import Enum, unique


@unique
class QuoteChangeKey(Enum):
    timestamp = 'timestamp'
    latest_price = 'latestPrice'
    prev_close = 'preClose'
    volume = 'volume'
    open = 'open'
    high = 'high'
    low = 'low'
    close = 'close'
    ask_price = 'askPrice'
    ask_size = 'askSize'
    bid_price = 'bidPrice'
    bid_size = 'bidSize'
    minute = 'mi'
    bid_depth = 'bidDepth'
    ask_depth = 'askDepth'


@unique
class QuoteKeyType(Enum):
    ALL = 'askPrice,askSize,bidPrice,bidSize,open,high,low,close,preClose,volume,latestPrice,mi'  # 所有行情数据
    QUOTE = 'askPrice,askSize,bidPrice,bidSize'  # 盘口数据
    TRADE = 'open,high,low,close,preClose,volume,latestPrice'  # 成交数据
    TIMELINE = 'mi'  # 分时数据



from functools import partial
from .client import trade_client, global_context, quote_client

create_order = partial(trade_client.create_order, account=global_context.account)
place_order = trade_client.place_order
get_open_orders = partial(trade_client.get_open_orders, account=global_context.account)


def _order(contract, action, order_type, quantity, limit_price=None, aux_price=None,
           trail_stop_price=None, trailing_percent=None, percent_offset=None, time_in_force=None,
           outside_rth=None):
    order_obj = create_order(contract=contract, action=action, order_type=order_type, quantity=quantity,
                             limit_price=limit_price, aux_price=aux_price, trail_stop_price=trail_stop_price,
                             trailing_percent=trailing_percent, percent_offset=percent_offset,
                             time_in_force=time_in_force, outside_rth=outside_rth)
    if order_obj and place_order(order_obj):
        return order_obj.order_id
    return False


def order(asset, amount, style='market', limit_price=0.0):
    if amount == 0:
        return
    action = 'BUY' if amount > 0 else 'SELL'
    contract = global_context.contract_map.get(asset)
    if style == 'market':
        return _order(contract=contract, action=action, order_type='MKT', quantity=abs(amount))
    elif style == 'limit' and limit_price > 0:
        return _order(contract=contract, action=action, order_type='LMT', quantity=abs(amount), limit_price=limit_price)


def order_value(asset, value, style='market', limit_price=0.0):
    action = 'BUY' if value > 0 else 'SELL'
    contract = global_context.contract_map.get(asset)
    if style == 'market':
        curr_close = quote_client.get_stock_briefs(symbols=[asset]).latest_price.iloc[0]
        if curr_close is None:
            raise Exception('no price data')
        return _order(contract=contract, action=action, order_type='MKT', quantity=abs(value) // curr_close)
    elif style == 'limit' and limit_price > 0:
        return _order(contract=contract, action=action, order_type='LMT', quantity=abs(value) // limit_price, limit_price=limit_price)


def get_order(order_id):
    return trade_client.get_order(account=global_context.account, order_id=order_id)


def cancel_order(order_id):
    return trade_client.cancel_order(account=global_context.account, order_id=order_id)


def order_percent(asset, percent, style='market', limit_price=0.0):
    raise NotImplemented
    # portfolio_value = global_context.asset_manager.summary.gross_position_value + global_context.asset_manager.summary.available_funds
    # value = portfolio_value * percent
    # return order_value(asset, value, style, limit_price)


def order_target(asset, amount, style='market', limit_price=0.0):
    raise NotImplemented
    # curr_position = global_context.position_manager.get(asset)
    # if curr_position:
    #     if curr_position.quantity != amount:
    #         diff_amount = amount - curr_position.quantity
    #         return order(asset, diff_amount, style, limit_price)
    # else:
    #     return order(asset, amount, style, limit_price)


def order_target_value(asset, amount, style='market', limit_price=0.0):
    raise NotImplemented
    # if style == 'market':
    #     #curr_price = minute_bar_util.get_curr_bar(assets=asset,  fields='close')
    #     quote_client.get_stock_briefs(symbols=[asset]).latest_price.iloc[0]
    # else:
    #     curr_price = limit_price
    # if curr_price > 0:
    #     target_qty = abs(amount) // curr_price
    #     return order_target(asset, target_qty if amount >= 0 else (-target_qty), style, limit_price)


def order_target_percent(asset, percent, style='market', limit_price=0.0):
    raise NotImplemented
    # target_value = global_context.asset_manager.summary.gross_position_value + global_context.asset_manager.summary.available_funds * percent
    # return order_target_value(asset, target_value, style, limit_price)





import json
from tigeropen.common.response import TigerResponse

PREVIEW_ORDER_FIELD_MAPPING = {"initMarginBefore": "init_margin_before", "commissionCurrency": "commission_currency",
                               "maintMargin": "maint_margin", "equityWithLoan": "equity_with_loan",
                               "minCommission": "min_commission", "maintMarginBefore": "maint_margin_before",
                               "initMargin": "init_margin", "equityWithLoanBefore": "equity_with_loan_before",
                               "marginCurrency": "margin_currency", "maxCommission": "max_commission",
                               "warningText": "warning_text"}


class PreviewOrderResponse(TigerResponse):
    def __init__(self):
        super(PreviewOrderResponse, self).__init__()
        self.preview_order = dict()
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(PreviewOrderResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data:
            data_json = json.loads(self.data)
            for key, value in data_json.items():
                field = PREVIEW_ORDER_FIELD_MAPPING.get(key, key)
                self.preview_order[field] = value




"""
provide compatibility for tiger quant platform strategies.
"""
import pytz
from datetime import datetime, timedelta
from tigeropen.common.consts import BarPeriod
from .context import global_context
from .client import quote_client
from .setting import MARKET
import logbook
logbook.set_datetime_format("local")
log = logbook.Logger('engine')


def symbol(symbol_str, currency=MARKET.CURRENCY, security_type='STK', exchange=None):
    """
    :param symbol_str: 证券类型, 默认股票
    :param security_type: 非股票需要指定security_type
    :param currency: 非美股symbol需要指定currency
    :param exchange:
    :return:
    """
    global_context.subscribe(symbol_str.upper(), currency=currency, security_type=security_type, exchange=exchange)
    return symbol_str


def symbols(*args, currency=MARKET.CURRENCY, security_type='STK', exchange=None):
    for sym in list(args):
        symbol(sym, currency=currency, security_type=security_type, exchange=exchange)
    return list(args)


def record(**kwargs):
    pass


def set_benchmark(symbol):
    pass


def schedule_function(func, date_rule=None, time_rule=None):
    global_context.schedule_function = Schedule(func, date_rule, time_rule)


class Schedule:
    def __init__(self, func, date_rule=None, time_rule=None):
        self.func = func
        self.date_rule = date_rule
        self.time_rule = time_rule
        self.last_trade_date = self.get_last_trade_date()
        self.market_end_time = MARKET.CLOSE_TIME - timedelta(minutes=2)

    def get_last_trade_date(self):
        ms_timestamp = quote_client.get_bars(symbols=[MARKET.INDEX], period=BarPeriod.DAY, limit=1).time.iloc[0]
        return datetime.fromtimestamp(ms_timestamp // 1000, tz=pytz.timezone(MARKET.TIMEZONE))

    def run(self, data):
        curr_time = datetime.now(tz=pytz.timezone(MARKET.TIMEZONE))
        curr_clock = curr_time.replace(second=0).strftime('%H:%M:%S')
        if (self.date_rule == 'month_start' and self.last_trade_date.month != curr_time.month) or self.date_rule is None:
            if (self.time_rule == 'market_open' or self.time_rule is None) and curr_clock == MARKET.OPEN_TIME:
                self.func(global_context, data)
            elif self.time_rule == 'market_end' and curr_clock == self.market_end_time:
                self.func(global_context, data)
            elif self.time_rule == 'always':
                self.func(global_context, data)


# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
import json
from tigeropen.common.consts import PYTHON_VERSION_3, THREAD_LOCAL
from tigeropen.common.exceptions import RequestException, ResponseException

try:
    import httplib
except ImportError:
    import http.client as httplib
try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse
try:
    from urllib.parse import quote_plus
except ImportError:
    from urllib import quote_plus


def url_encode(params, charset):
    query_string = ""
    for (k, v) in params.items():
        value = v
        if not isinstance(value, str):
            value = json.dumps(value, ensure_ascii=False)
        if PYTHON_VERSION_3:
            value = quote_plus(value, encoding=charset)
        else:
            value = quote_plus(value)
        query_string += ("&" + k + "=" + value)
    query_string = query_string[1:]
    return query_string


def get_http_connection(url, query_string, timeout):
    url_parse_result = urlparse.urlparse(url)
    host = url_parse_result.hostname
    port = 80
    connection = httplib.HTTPConnection(host=host, port=port, timeout=timeout)
    if url.find("https") == 0:
        port = 443
        connection = httplib.HTTPSConnection(host=host, port=port, timeout=timeout)
    url = url_parse_result.scheme + "://" + url_parse_result.hostname
    if url_parse_result.port:
        url += url_parse_result.port
    url += url_parse_result.path
    if query_string:
        url += ('?' + query_string)
    return url, connection


def do_post(url, query_string=None, headers=None, params=None, timeout=15, charset=None):
    url, connection = get_http_connection(url, query_string, timeout)

    try:
        connection.connect()
    except Exception as e:
        raise RequestException('[' + THREAD_LOCAL.uuid + ']post connect failed. ' + str(e))
    try:
        connection.request("POST", url, body=json.dumps(params), headers=headers)
    except Exception as e:
        raise RequestException('[' + THREAD_LOCAL.uuid + ']post request failed. ' + str(e))
    response = connection.getresponse()
    result = response.read()

    if response.status != 200:
        if PYTHON_VERSION_3 and charset:
            result = result.decode(charset)
        raise ResponseException('[' + THREAD_LOCAL.uuid + ']invalid http status ' + str(response.status) +
                                ',detail body:' + result)
    try:
        response.close()
        connection.close()
    except Exception as e:
        pass
    return result



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import six
import pandas as pd
from tigeropen.common.response import TigerResponse
from tigeropen.common.util.string_utils import get_string
from tigeropen.common.util.common_utils import eastern
from tigeropen.common.util.contract_utils import get_option_identifier

COLUMNS = ['identifier', 'symbol', 'expiry', 'strike', 'put_call', 'multiplier', 'ask_price', 'ask_size', 'bid_price',
           'bid_size', 'pre_close', 'latest_price', 'latest_time', 'volume', 'open_interest', 'open', 'high', 'low',
           'rates_bonds', 'volatility']
BRIEF_FIELD_MAPPINGS = {'askPrice': 'ask_price', 'askSize': 'ask_size', 'bidPrice': 'bid_price', 'bidSize': 'bid_size',
                        'latestPrice': 'latest_price', 'openInterest': 'open_interest', 'preClose': 'pre_close',
                        'right': 'put_call', 'latestTime': 'latest_time', 'openInt': 'open_interest',
                        'ratesBonds': 'rates_bonds'}


class OptionBriefsResponse(TigerResponse):
    def __init__(self):
        super(OptionBriefsResponse, self).__init__()
        self.briefs = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(OptionBriefsResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            brief_data = []
            for item in self.data:
                item_values = {}
                for key, value in item.items():
                    if value is None:
                        continue
                    if isinstance(value, six.string_types):
                        value = get_string(value)
                    if key == 'right':
                        value = value.upper()
                    tag = BRIEF_FIELD_MAPPINGS[key] if key in BRIEF_FIELD_MAPPINGS else key
                    item_values[tag] = value
                if 'identifier' not in item_values:
                    underlying_symbol = item_values.get('symbol')
                    expiry = item_values.get('expiry')
                    strike = float(item_values.get('strike'))
                    put_call = item_values.get('right')
                    expiry = pd.Timestamp(expiry, unit='ms', tzinfo=eastern).date().strftime("%Y%m%d")
                    item_values['identifier'] = get_option_identifier(underlying_symbol, expiry, put_call, strike)

                brief_data.append([item_values.get(tag) for tag in COLUMNS])

            self.briefs = pd.DataFrame(brief_data, columns=COLUMNS)



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import six

import pandas as pd
from tigeropen.common.util.string_utils import get_string
from tigeropen.common.response import TigerResponse

COLUMNS = ['symbol', 'settlement_date', 'short_interest', 'avg_daily_volume', 'days_to_cover', 'percent_of_float']
SHORT_INTEREST_FIELD_MAPPINGS = {'settlementDate': 'settlement_date', 'shortInterest': 'short_interest',
                                 'avgDailyVolume': 'avg_daily_volume', 'daysToCover': 'days_to_cover',
                                 'percentOfFloat': 'percent_of_float'}


class ShortInterestResponse(TigerResponse):
    def __init__(self):
        super(ShortInterestResponse, self).__init__()
        self.short_interests = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(ShortInterestResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            short_interest_items = []
            for symbol_item in self.data:
                symbol = symbol_item.get('symbol')
                items = symbol_item.get('items')
                for item in items:
                    item_values = {'symbol': symbol}
                    for key, value in item.items():
                        if value is None:
                            continue
                        if isinstance(value, six.string_types):
                            value = get_string(value)
                        tag = SHORT_INTEREST_FIELD_MAPPINGS[key] if key in SHORT_INTEREST_FIELD_MAPPINGS else key
                        item_values[tag] = value
                    short_interest_items.append([item_values.get(tag) for tag in COLUMNS])

            self.short_interests = pd.DataFrame(short_interest_items, columns=COLUMNS)



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import pandas as pd
from tigeropen.common.response import TigerResponse


class OptionExpirationsResponse(TigerResponse):
    def __init__(self):
        super(OptionExpirationsResponse, self).__init__()
        self.expirations = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(OptionExpirationsResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            expiration_items = []
            for item in self.data:
                symbol = item.get('symbol')
                dates = item.get('dates')
                timestamps = item.get('timestamps')

                if symbol and dates:
                    for date, timestamp in zip(dates, timestamps):
                        expiration_items.append([symbol, date, timestamp])

            self.expirations = pd.DataFrame(expiration_items, columns=['symbol', 'date', 'timestamp'])



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import json
from tigeropen.common.response import TigerResponse
from tigeropen.common.util.string_utils import get_string
from tigeropen.trade.domain.profile import AccountProfile


class ProfilesResponse(TigerResponse):
    def __init__(self):
        super(ProfilesResponse, self).__init__()
        self.profiles = []
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(ProfilesResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data:
            data_json = json.loads(self.data)
            if 'items' in data_json:
                for item in data_json['items']:
                    account, capability, status = None, None, None
                    for key, value in item.items():
                        if value is None:
                            continue
                        if key == 'account':
                            account = get_string(value)
                        elif key == 'capability':
                            capability = get_string(value)
                        elif key == 'status':
                            status = get_string(value)
                    profile = AccountProfile(account, capability, status)
                    self.profiles.append(profile)



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""


class MarketStatus(object):
    def __init__(self, market, status, open_time, trading_status):
        self.market = market
        self.status = status
        self.open_time = open_time
        self.trading_status = trading_status

    def __repr__(self):
        """
        String representation for this object.
        """
        return "MarketStatus(%s)" % self.__dict__



# -*- coding: utf-8 -*-

from enum import Enum, unique


@unique
class Valuation(Enum):
    """
    估值指标
    tev: total enterprice value
    bv: book value
    tangbv: tangible book value
    fcf: free cashflow
    excl: exclued
    extra: extraordinary
    """
    tev_to_ltm_total_revenues = "tev_to_ltm_total_revenues"
    tev_to_ltm_ebitda = "tev_to_ltm_ebitda"
    tev_to_ltm_ebit = "tev_to_ltm_ebit"
    p_to_ltm_diluted_eps_before_extra = "p_to_ltm_diluted_eps_before_extra"
    p_to_bv = "p_to_bv"
    p_to_tangbv = "p_to_tangbv"
    market_cap_to_ltm_total_revenues = "market_cap_to_ltm_total_revenues"
    market_cap_to_ltm_ebt_excl_unusual_items = "market_cap_to_ltm_ebt_excl_unusual_items"
    tev_to_ltm_unlevered_fcf = "tev_to_ltm_unlevered_fcf"
    market_cap_to_ltm_levered_fcf = "market_cap_to_ltm_levered_fcf"
    total_enterprise_value = "total_enterprise_value"
    market_capitalization = "market_capitalization"
    shares_outstanding = "shares_outstanding"


@unique
class Income(Enum):
    """
    利润表
    """
    # gross profit
    total_revenue = "total_revenue"
    revenues = "revenues"
    other_revenues_total = "other_revenues_total"
    finance_div_revenues = "finance_div_revenues"
    insurance_div_revenues = "insurance_div_revenues"
    gain_loss_on_sale_of_assets_rev = "gain_loss_on_sale_of_assets_rev"
    gain_loss_on_sale_of_invest_rev = "gain_loss_on_sale_of_invest_rev"
    interest_and_invest_income_rev = "interest_and_invest_income_rev"
    other_revenues = "other_revenues"
    cost_of_revenues = "cost_of_revenues"
    cost_of_goods_sold = "cost_of_goods_sold"
    finance_div_operating_exp = "finance_div_operating_exp"
    insurance_div_operating_exp = "insurance_div_operating_exp"
    interest_expensefinance_division = "interest_expensefinance_division"
    gross_profit = "gross_profit"

    # operating income
    sga_exp_total = "sga_exp_total"
    selling_general_admin_exp = "selling_general_admin_exp"
    explorationdrilling_costs = "explorationdrilling_costs"
    provision_for_bad_debts = "provision_for_bad_debts"
    stock_based_compensation = "stock_based_compensation"
    pre_opening_costs = "pre_opening_costs"
    rd_exp = "rd_exp"
    depreciation_amort_total = "depreciation_amort_total"
    depreciation_amort = "depreciation_amort"
    amort_of_goodwill_and_intangibles = "amort_of_goodwill_and_intangibles"
    impair_of_oil_gas_mineral_prop = "impair_of_oil_gas_mineral_prop"
    other_operating_expense_income = "other_operating_expense_income"
    other_operating_exp_total = "other_operating_exp_total"
    total_operating_expenses = "total_operating_expenses"
    operating_income = "operating_income"

    # net income
    interest_expense = "interest_expense"
    interest_and_invest_income = "interest_and_invest_income"
    interest_and_dividend_income = "interest_and_dividend_income"
    investment_income = "investment_income"
    net_interest_exp = "net_interest_exp"
    other_non_operating_exp_total = "other_non_operating_exp_total"
    income_loss_from_affiliates = "income_loss_from_affiliates"
    currency_exchange_gains_loss = "currency_exchange_gains_loss"
    other_non_operating_inc_exp = "other_non_operating_inc_exp"
    ebt_excl_unusual_items = "ebt_excl_unusual_items"
    merger_restruct_charges = "merger_restruct_charges"
    restructuring_charges = "restructuring_charges"
    merger_related_restruct_charges = "merger_related_restruct_charges"
    impairment_of_goodwill = "impairment_of_goodwill"
    gain_loss_on_sale_of_invest = "gain_loss_on_sale_of_invest"
    gain_loss_on_sale_of_assets = "gain_loss_on_sale_of_assets"
    other_unusual_items_total = "other_unusual_items_total"
    asset_writedown = "asset_writedown"
    in_process_rd_exp = "in_process_rd_exp"
    insurance_settlements = "insurance_settlements"
    legal_settlements = "legal_settlements"
    other_unusual_items = "other_unusual_items"
    total_unusual_items = "total_unusual_items"
    ebt_incl_unusual_items = "ebt_incl_unusual_items"
    income_tax_expense = "income_tax_expense"
    earnings_from_cont_ops = "earnings_from_cont_ops"
    earnings_of_discontinued_ops = "earnings_of_discontinued_ops"
    extraord_item_account_change = "extraord_item_account_change"
    other_expenses_including_taxes_private_only = "other_expenses_including_taxes_private_only"
    net_income_to_company = "net_income_to_company"
    minority_int_in_earnings = "minority_int_in_earnings"
    net_income = "net_income"
    pref_dividends_and_other_adj = "pref_dividends_and_other_adj"
    ni_to_common_incl_extra_items = "ni_to_common_incl_extra_items"
    ni_to_common_excl_extra_items = "ni_to_common_excl_extra_items"

    # per share items
    basic_eps = "basic_eps"
    basic_eps_excl_extra_items = "basic_eps_excl_extra_items"
    weighted_avg_shares_outstanding = "weighted_avg_shares_outstanding"
    diluted_eps_incl_extra_items = "diluted_eps_incl_extra_items"
    diluted_eps_excl_extra_items = "diluted_eps_excl_extra_items"
    weighted_avg_diluted_shares_out = "weighted_avg_diluted_shares_out"
    normalized_basic_eps = "normalized_basic_eps"
    normalized_diluted_eps = "normalized_diluted_eps"
    dividends_per_share = "dividends_per_share"
    gross_dividends_per_share = "gross_dividends_per_share"
    net_dividends_per_share = "net_dividends_per_share"
    special_gross_dividends_per_share = "special_gross_dividends_per_share"
    special_net_dividends_per_share = "special_net_dividends_per_share"

    # supplemental items
    corporate_tax_rate = "corporate_tax_rate"
    imputed_dividend_amount = "imputed_dividend_amount"
    imputation_percentage = "imputation_percentage"
    imputation_credit_amount = "imputation_credit_amount"
    special_dividend_imputed_dividend_amount = "special_dividend_imputed_dividend_amount"
    special_dividend_imputation_percentage = "special_dividend_imputation_percentage"
    special_dividend_imputation_credit_amount = "special_dividend_imputation_credit_amount"
    payout_ratio = "payout_ratio"
    ebitda = "ebitda"
    ebita = "ebita"
    ebit = "ebit"
    ebitdar = "ebitdar"
    ebitda_incl_equity_inc_from_affiliates = "ebitda_incl_equity_inc_from_affiliates"
    ebita_incl_equity_inc_from_affiliates = "ebita_incl_equity_inc_from_affiliates"
    ebit_incl_equity_inc_from_affiliates = "ebit_incl_equity_inc_from_affiliates"
    ebitdar_incl_equity_inc_from_affiliates = "ebitdar_incl_equity_inc_from_affiliates"
    ebitda_excl_sbc_and_incl_equity_inc_from_affiliates = "ebitda_excl_sbc_and_incl_equity_inc_from_affiliates"
    ebita_excl_sbc_and_incl_equity_inc_from_affiliates = "ebita_excl_sbc_and_incl_equity_inc_from_affiliates"
    ebit_excl_sbc_and_incl_equity_inc_from_affiliates = "ebit_excl_sbc_and_incl_equity_inc_from_affiliates"
    ebitdar_excl_sbc_and_incl_equity_inc_from_affiliates = "ebitdar_excl_sbc_and_incl_equity_inc_from_affiliates"
    ebitda_excl_sbc = "ebitda_excl_sbc"
    ebita_excl_sbc = "ebita_excl_sbc"
    ebit_excl_sbc = "ebit_excl_sbc"
    ebitdar_excl_sbc = "ebitdar_excl_sbc"
    ebitdacapex = "ebitdacapex"
    effective_tax_rate = "effective_tax_rate"
    normalized_net_income = "normalized_net_income"
    owned_operated_same_store_sales_growth = "owned_operated_same_store_sales_growth"
    as_reported_total_revenue = "as_reported_total_revenue"
    provision_for_doubtful_accountspatient_service_revenue = "provision_for_doubtful_accountspatient_service_revenue"
    ni_per_sfas_123_after_options = "ni_per_sfas_123_after_options"
    advertising_expense = "advertising_expense"
    marketing_expense = "marketing_expense"
    selling_and_marketing_expense = "selling_and_marketing_expense"
    general_and_administrative_expense = "general_and_administrative_expense"
    rd_expense = "rd_expense"
    net_rental_expense_total = "net_rental_expense_total"
    minimum_rental_expense = "minimum_rental_expense"
    depreciation_of_rental_assets = "depreciation_of_rental_assets"
    revenues_per_share = "revenues_per_share"
    total_revenues_employee = "total_revenues_employee"
    diluted_net_income = "diluted_net_income"
    imputed_oper_lease_interest_exp = "imputed_oper_lease_interest_exp"
    imputed_oper_lease_depreciation = "imputed_oper_lease_depreciation"
    stock_based_comp_cogs = "stock_based_comp_cogs"
    stock_based_comp_rd_exp = "stock_based_comp_rd_exp"
    stock_based_comp_sm_exp = "stock_based_comp_sm_exp"
    stock_based_comp_ga_exp = "stock_based_comp_ga_exp"
    stock_based_comp_sga_exp = "stock_based_comp_sga_exp"
    stock_based_comp_unallocated = "stock_based_comp_unallocated"
    stock_based_comp_after_tax = "stock_based_comp_after_tax"
    stock_based_comp_total = "stock_based_comp_total"
    non_cash_pension_expense = "non_cash_pension_expense"
    distributable_cash = "distributable_cash"
    standardized_distributable_cash = "standardized_distributable_cash"
    distributable_cash_per_share = "distributable_cash_per_share"
    annualized_distributions_per_unit = "annualized_distributions_per_unit"
    distributable_cash_payout_ratio_income_trusts = "distributable_cash_payout_ratio_income_trusts"
    debt_issuance_costs = "debt_issuance_costs"
    period_date_income_statement = "period_date_income_statement"
    filing_date_income_statement = "filing_date_income_statement"
    financial_accounting_standard = "financial_accounting_standard"
    filing_currency = "filing_currency"


@unique
class Balance(Enum):
    """
    资产负债表
    """
    # assets
    cash_and_equivalents = "cash_and_equivalents"
    short_term_investments = "short_term_investments"
    trading_asset_securities = "trading_asset_securities"
    total_cash_st_investments = "total_cash_st_investments"
    accounts_receivable = "accounts_receivable"
    other_receivables = "other_receivables"
    notes_receivable = "notes_receivable"
    total_receivables = "total_receivables"
    inventory = "inventory"
    prepaid_exp = "prepaid_exp"
    finance_div_loans_and_leases_st = "finance_div_loans_and_leases_st"
    finance_div_other_curr_assets = "finance_div_other_curr_assets"
    other_current_assets_total = "other_current_assets_total"
    loans_held_for_sale = "loans_held_for_sale"
    deferred_tax_assets_curr = "deferred_tax_assets_curr"
    restricted_cash = "restricted_cash"
    other_current_assets = "other_current_assets"
    total_current_assets = "total_current_assets"
    gross_property_plant_equipment = "gross_property_plant_equipment"
    accumulated_depreciation = "accumulated_depreciation"
    net_property_plant_equipment = "net_property_plant_equipment"
    longterm_investments = "longterm_investments"
    goodwill = "goodwill"
    other_intangibles = "other_intangibles"
    finance_div_loans_and_leases_lt = "finance_div_loans_and_leases_lt"
    finance_div_other_lt_assets = "finance_div_other_lt_assets"
    other_assets_total = "other_assets_total"
    capitalizedpurchased_software = "capitalizedpurchased_software"
    accounts_receivable_longterm = "accounts_receivable_longterm"
    loans_receivable_longterm = "loans_receivable_longterm"
    deferred_tax_assets_lt = "deferred_tax_assets_lt"
    deferred_charges_lt = "deferred_charges_lt"
    other_longterm_assets = "other_longterm_assets"
    total_assets = "total_assets"

    # liabilities
    accounts_payable = "accounts_payable"
    accrued_exp = "accrued_exp"
    total_shortterm_borrowings = "total_shortterm_borrowings"
    curr_port_of_lt_debtcap_leases = "curr_port_of_lt_debtcap_leases"
    curr_port_of_long_term_debt = "curr_port_of_long_term_debt"
    curr_port_of_cap_leases = "curr_port_of_cap_leases"
    finance_div_debt_current = "finance_div_debt_current"
    finance_div_other_curr_liab = "finance_div_other_curr_liab"
    other_current_liabilities_total = "other_current_liabilities_total"
    curr_income_taxes_payable = "curr_income_taxes_payable"
    unearned_revenue_current = "unearned_revenue_current"
    def_tax_liability_curr = "def_tax_liability_curr"
    other_current_liabilities = "other_current_liabilities"
    total_current_liabilities = "total_current_liabilities"
    longterm_debt = "longterm_debt"
    capital_leases = "capital_leases"
    finance_div_debt_noncurr = "finance_div_debt_noncurr"
    finance_div_other_noncurr_liab = "finance_div_other_noncurr_liab"
    other_liabilities_total = "other_liabilities_total"
    unearned_revenue_noncurrent = "unearned_revenue_noncurrent"
    pension_other_postretire_benefits = "pension_other_postretire_benefits"
    def_tax_liability_noncurr = "def_tax_liability_noncurr"
    other_noncurrent_liabilities = "other_noncurrent_liabilities"
    total_liabilities = "total_liabilities"

    # equity
    pref_stock_redeemable = "pref_stock_redeemable"
    pref_stock_nonredeem = "pref_stock_nonredeem"
    pref_stock_convertible = "pref_stock_convertible"
    pref_stock_other = "pref_stock_other"
    common_stock = "common_stock"
    total_pref_equity = "total_pref_equity"
    common_stock_apic = "common_stock_apic"
    additional_paid_in_capital = "additional_paid_in_capital"
    retained_earnings = "retained_earnings"
    treasury_stock_other = "treasury_stock_other"
    treasury_stock = "treasury_stock"
    comprehensive_inc_and_other = "comprehensive_inc_and_other"
    total_common_equity = "total_common_equity"
    minority_interest = "minority_interest"
    total_equity = "total_equity"
    total_liabilities_and_equity = "total_liabilities_and_equity"

    # supplemental
    number_of_equity_capital_structure_share_classes = "number_of_equity_capital_structure_share_classes"
    total_shares_out_on_filing_date = "total_shares_out_on_filing_date"
    total_shares_out_on_balance_sheet_date = "total_shares_out_on_balance_sheet_date"
    book_valueshare = "book_valueshare"
    book_valueshare_reported = "book_valueshare_reported"
    tangible_book_value = "tangible_book_value"
    tangible_book_valueshare = "tangible_book_valueshare"
    tangible_book_valueshare_reported = "tangible_book_valueshare_reported"
    total_intangibles = "total_intangibles"
    total_debt = "total_debt"
    total_current_debt = "total_current_debt"
    total_noncurrent_debt = "total_noncurrent_debt"
    net_debt = "net_debt"
    total_redeemable_minority_interest = "total_redeemable_minority_interest"
    total_minority_interest = "total_minority_interest"
    cash_per_share = "cash_per_share"
    total_capital = "total_capital"
    working_capital = "working_capital"
    net_working_capital = "net_working_capital"
    liquidation_value_of_preferred_stock_convertible = "liquidation_value_of_preferred_stock_convertible"
    liquidation_value_of_preferred_stock_nonredeemable = "liquidation_value_of_preferred_stock_nonredeemable"
    liquidation_value_of_preferred_stock_redeemable = "liquidation_value_of_preferred_stock_redeemable"
    buildings = "buildings"
    land = "land"
    lifo_reserve = "lifo_reserve"
    machinery = "machinery"
    employees_under_union_contracts = "employees_under_union_contracts"
    parttime_employees = "parttime_employees"
    total_employees = "total_employees"
    equity_method_investments = "equity_method_investments"
    finished_goods_inventory = "finished_goods_inventory"
    full_time_employees = "full_time_employees"
    raw_materials_inventory = "raw_materials_inventory"
    work_in_progress_inventory = "work_in_progress_inventory"
    construction_in_progress = "construction_in_progress"
    debt_equivalent_oper_leases = "debt_equivalent_oper_leases"
    debt_equivalent_of_unfunded_pbo = "debt_equivalent_of_unfunded_pbo"
    cost_of_borrowing = "cost_of_borrowing"
    accounts_receivables_unbilled = "accounts_receivables_unbilled"
    period_date_balance_sheet = "period_date_balance_sheet"
    filing_date_balance_sheet = "filing_date_balance_sheet"


@unique
class CashFlow(Enum):
    """
    现金流量表
    """
    # operating cashflow
    net_income_cf = "net_income_cf"
    depreciation_amort_cf = "depreciation_amort_cf"
    amort_of_goodwill_and_intangibles_cf = "amort_of_goodwill_and_intangibles_cf"
    impair_of_oil_gas_mineral_prop_cf = "impair_of_oil_gas_mineral_prop_cf"
    depreciation_amort_total_cf = "depreciation_amort_total_cf"
    other_amortization = "other_amortization"
    other_noncash_items_total = "other_noncash_items_total"
    minority_int_in_earnings_cf = "minority_int_in_earnings_cf"
    gain_loss_from_sale_of_asset = "gain_loss_from_sale_of_asset"
    gain_loss_on_sale_of_invest_cf = "gain_loss_on_sale_of_invest_cf"
    asset_writedown_restructuring_costs = "asset_writedown_restructuring_costs"
    net_increase_decrease_in_loans_origsold = "net_increase_decrease_in_loans_origsold"
    provision_for_credit_losses = "provision_for_credit_losses"
    income_loss_on_equity_invest = "income_loss_on_equity_invest"
    stockbased_compensation = "stockbased_compensation"
    tax_benefit_from_stock_options = "tax_benefit_from_stock_options"
    provision_writeoff_of_bad_debts = "provision_writeoff_of_bad_debts"
    net_cash_from_discontinued_ops = "net_cash_from_discontinued_ops"
    other_operating_activities = "other_operating_activities"
    change_in_net_operating_assets = "change_in_net_operating_assets"
    change_in_trad_asset_securities = "change_in_trad_asset_securities"
    change_in_accounts_receivable = "change_in_accounts_receivable"
    change_in_inventories = "change_in_inventories"
    change_in_acc_payable = "change_in_acc_payable"
    change_in_unearned_rev = "change_in_unearned_rev"
    change_in_inc_taxes = "change_in_inc_taxes"
    change_in_def_taxes = "change_in_def_taxes"
    change_in_other_net_operating_assets = "change_in_other_net_operating_assets"
    cash_from_ops = "cash_from_ops"

    # investing cashflow
    capital_expenditure = "capital_expenditure"
    sale_of_property_plant_and_equipment = "sale_of_property_plant_and_equipment"
    cash_acquisitions = "cash_acquisitions"
    divestitures = "divestitures"
    other_investing_activities_total = "other_investing_activities_total"
    sale_purchase_of_real_estate_properties = "sale_purchase_of_real_estate_properties"
    sale_purchase_of_intangible_assets = "sale_purchase_of_intangible_assets"
    invest_in_marketable_equity_securt = "invest_in_marketable_equity_securt"
    net_inc_dec_in_loans_originatedsold = "net_inc_dec_in_loans_originatedsold"
    other_investing_activities = "other_investing_activities"
    cash_from_investing = "cash_from_investing"

    # financing cashflow
    short_term_debt_issued = "short_term_debt_issued"
    longterm_debt_issued = "longterm_debt_issued"
    total_debt_issued = "total_debt_issued"
    short_term_debt_repaid = "short_term_debt_repaid"
    longterm_debt_repaid = "longterm_debt_repaid"
    total_debt_repaid = "total_debt_repaid"
    issuance_of_common_stock = "issuance_of_common_stock"
    repurchase_of_common_stock = "repurchase_of_common_stock"
    issuance_of_preferred_stock = "issuance_of_preferred_stock"
    repurchase_of_preferred_stock = "repurchase_of_preferred_stock"
    common_dividends_paid = "common_dividends_paid"
    pref_dividends_paid = "pref_dividends_paid"
    common_andor_pref_dividends_paid = "common_andor_pref_dividends_paid"
    total_dividends_paid = "total_dividends_paid"
    other_financing_activities_total = "other_financing_activities_total"
    special_dividend_paid = "special_dividend_paid"
    other_financing_activities = "other_financing_activities"
    cash_from_financing = "cash_from_financing"

    # summary
    foreign_exchange_rate_adj = "foreign_exchange_rate_adj"
    misc_cash_flow_adj = "misc_cash_flow_adj"
    net_change_in_cash = "net_change_in_cash"

    # supplementals
    cash_interest_paid = "cash_interest_paid"
    cash_taxes_paid = "cash_taxes_paid"
    levered_free_cash_flow = "levered_free_cash_flow"
    unlevered_free_cash_flow = "unlevered_free_cash_flow"
    cash_flow_per_share = "cash_flow_per_share"
    change_in_net_working_capital = "change_in_net_working_capital"
    net_debt_issued = "net_debt_issued"
    maintenance_capex = "maintenance_capex"
    depreciation_of_rental_assets_cf = "depreciation_of_rental_assets_cf"
    sale_proceeds_from_rental_assets = "sale_proceeds_from_rental_assets"
    period_date_cash_flow = "period_date_cash_flow"
    filing_date_cash_flow = "filing_date_cash_flow"


@unique
class BalanceSheetRatio(Enum):
    """
    资产负债表相关比率
    """
    total_asset_turnover = "total_asset_turnover"
    fixed_asset_turnover = "fixed_asset_turnover"
    accounts_receivable_turnover = "accounts_receivable_turnover"
    inventory_turnover = "inventory_turnover"
    current_ratio = "current_ratio"
    quick_ratio = "quick_ratio"
    avg_days_sales_out = "avg_days_sales_out"
    avg_days_inventory_out = "avg_days_inventory_out"
    avg_days_payable_out = "avg_days_payable_out"
    avg_cash_conversion_cycle = "avg_cash_conversion_cycle"


@unique
class Growth(Enum):
    """
    成长能力
    """
    total_revenues_1_yr_growth = "total_revenues_1_yr_growth"
    gross_profit_1_yr_growth = "gross_profit_1_yr_growth"
    ebitda_1_yr_growth = "ebitda_1_yr_growth"
    ebita_1_yr_growth = "ebita_1_yr_growth"
    ebit_1_yr_growth = "ebit_1_yr_growth"
    earnings_from_cont_ops_1_yr_growth = "earnings_from_cont_ops_1_yr_growth"
    net_income_1_yr_growth = "net_income_1_yr_growth"
    normalized_net_income_1_yr_growth = "normalized_net_income_1_yr_growth"
    diluted_eps_before_extra_1_yr_growth = "diluted_eps_before_extra_1_yr_growth"
    accounts_receivable_1_yr_growth = "accounts_receivable_1_yr_growth"
    inventory_1_yr_growth = "inventory_1_yr_growth"
    net_ppe_1_yr_growth = "net_ppe_1_yr_growth"
    common_equity_1_yr_growth = "common_equity_1_yr_growth"
    total_assets_1_yr_growth = "total_assets_1_yr_growth"
    tangible_book_value_1_yr_growth = "tangible_book_value_1_yr_growth"
    cash_from_operations_1_yr_growth = "cash_from_operations_1_yr_growth"
    capital_expenditures_1_yr_growth = "capital_expenditures_1_yr_growth"
    levered_free_cash_flow_1_yr_growth = "levered_free_cash_flow_1_yr_growth"
    unlevered_free_cash_flow_1_yr_growth = "unlevered_free_cash_flow_1_yr_growth"
    dividend_per_share_1_yr_growth = "dividend_per_share_1_yr_growth"
    total_revenues_2_yr_cagr = "total_revenues_2_yr_cagr"
    gross_profit_2_yr_cagr = "gross_profit_2_yr_cagr"
    ebitda_2_yr_cagr = "ebitda_2_yr_cagr"
    ebita_2_yr_cagr = "ebita_2_yr_cagr"
    ebit_2_yr_cagr = "ebit_2_yr_cagr"
    earnings_from_cont_ops_2_yr_cagr = "earnings_from_cont_ops_2_yr_cagr"
    net_income_2_yr_cagr = "net_income_2_yr_cagr"
    normalized_net_income_2_yr_cagr = "normalized_net_income_2_yr_cagr"
    diluted_eps_before_extra_2_yr_cagr = "diluted_eps_before_extra_2_yr_cagr"
    accounts_receivable_2_yr_cagr = "accounts_receivable_2_yr_cagr"
    inventory_2_yr_cagr = "inventory_2_yr_cagr"
    net_ppe_2_yr_cagr = "net_ppe_2_yr_cagr"
    common_equity_2_yr_cagr = "common_equity_2_yr_cagr"
    total_assets_2_yr_cagr = "total_assets_2_yr_cagr"
    tangible_book_value_2_yr_cagr = "tangible_book_value_2_yr_cagr"
    cash_from_ops_2_yr_cagr = "cash_from_ops_2_yr_cagr"
    capital_expenditures_2_yr_cagr = "capital_expenditures_2_yr_cagr"
    levered_free_cash_flow_2_yr_cagr = "levered_free_cash_flow_2_yr_cagr"
    unlevered_free_cash_flow_2_yr_cagr = "unlevered_free_cash_flow_2_yr_cagr"
    dividend_per_share_2_yr_cagr = "dividend_per_share_2_yr_cagr"
    total_revenues_3_yr_cagr = "total_revenues_3_yr_cagr"
    gross_profit_3_yr_cagr = "gross_profit_3_yr_cagr"
    ebitda_3_yr_cagr = "ebitda_3_yr_cagr"
    ebita_3_yr_cagr = "ebita_3_yr_cagr"
    ebit_3_yr_cagr = "ebit_3_yr_cagr"
    earnings_from_cont_ops_3_yr_cagr = "earnings_from_cont_ops_3_yr_cagr"
    net_income_3_yr_cagr = "net_income_3_yr_cagr"
    normalized_net_income_3_yr_cagr = "normalized_net_income_3_yr_cagr"
    diluted_eps_before_extra_3_yr_cagr = "diluted_eps_before_extra_3_yr_cagr"
    accounts_receivable_3_yr_cagr = "accounts_receivable_3_yr_cagr"
    inventory_3_yr_cagr = "inventory_3_yr_cagr"
    net_ppe_3_yr_cagr = "net_ppe_3_yr_cagr"
    common_equity_3_yr_cagr = "common_equity_3_yr_cagr"
    total_assets_3_yr_cagr = "total_assets_3_yr_cagr"
    tangible_book_value_3_yr_cagr = "tangible_book_value_3_yr_cagr"
    cash_from_ops_3_yr_cagr = "cash_from_ops_3_yr_cagr"
    capital_expenditures_3_yr_cagr = "capital_expenditures_3_yr_cagr"
    levered_free_cash_flow_3_yr_cagr = "levered_free_cash_flow_3_yr_cagr"
    unlevered_free_cash_flow_3_yr_cagr = "unlevered_free_cash_flow_3_yr_cagr"
    dividend_per_share_3_yr_cagr = "dividend_per_share_3_yr_cagr"
    total_revenues_5_yr_cagr = "total_revenues_5_yr_cagr"
    gross_profit_5_yr_cagr = "gross_profit_5_yr_cagr"
    ebitda_5_yr_cagr = "ebitda_5_yr_cagr"
    ebita_5_yr_cagr = "ebita_5_yr_cagr"
    ebit_5_yr_cagr = "ebit_5_yr_cagr"
    earnings_from_cont_ops_5_yr_cagr = "earnings_from_cont_ops_5_yr_cagr"
    net_income_5_yr_cagr = "net_income_5_yr_cagr"
    normalized_net_income_5_yr_cagr = "normalized_net_income_5_yr_cagr"
    diluted_eps_before_extra_5_yr_cagr = "diluted_eps_before_extra_5_yr_cagr"
    accounts_receivable_5_yr_cagr = "accounts_receivable_5_yr_cagr"
    inventory_5_yr_cagr = "inventory_5_yr_cagr"
    net_ppe_5_yr_cagr = "net_ppe_5_yr_cagr"
    common_equity_5_yr_cagr = "common_equity_5_yr_cagr"
    total_assets_5_yr_cagr = "total_assets_5_yr_cagr"
    tangible_book_value_5_yr_cagr = "tangible_book_value_5_yr_cagr"
    cash_from_ops_5_yr_cagr = "cash_from_ops_5_yr_cagr"
    capital_expenditures_5_yr_cagr = "capital_expenditures_5_yr_cagr"
    levered_free_cash_flow_5_yr_cagr = "levered_free_cash_flow_5_yr_cagr"
    unlevered_free_cash_flow_5_yr_cagr = "unlevered_free_cash_flow_5_yr_cagr"
    dividend_per_share_5_yr_cagr = "dividend_per_share_5_yr_cagr"
    total_revenues_7_yr_cagr = "total_revenues_7_yr_cagr"
    gross_profit_7_yr_cagr = "gross_profit_7_yr_cagr"
    ebitda_7_yr_cagr = "ebitda_7_yr_cagr"
    ebita_7_yr_cagr = "ebita_7_yr_cagr"
    ebit_7_yr_cagr = "ebit_7_yr_cagr"
    earnings_from_cont_ops_7_yr_cagr = "earnings_from_cont_ops_7_yr_cagr"
    net_income_7_yr_cagr = "net_income_7_yr_cagr"
    normalized_net_income_7_yr_cagr = "normalized_net_income_7_yr_cagr"
    diluted_eps_before_extra_7_yr_cagr = "diluted_eps_before_extra_7_yr_cagr"
    accounts_receivable_7_yr_cagr = "accounts_receivable_7_yr_cagr"
    inventory_7_yr_cagr = "inventory_7_yr_cagr"
    net_ppe_7_yr_cagr = "net_ppe_7_yr_cagr"
    common_equity_7_yr_cagr = "common_equity_7_yr_cagr"
    total_assets_7_yr_cagr = "total_assets_7_yr_cagr"
    tangible_book_value_7_yr_cagr = "tangible_book_value_7_yr_cagr"
    cash_from_ops_7_yr_cagr = "cash_from_ops_7_yr_cagr"
    capital_expenditures_7_yr_cagr = "capital_expenditures_7_yr_cagr"
    levered_free_cash_flow_7_yr_cagr = "levered_free_cash_flow_7_yr_cagr"
    unlevered_free_cash_flow_7_yr_cagr = "unlevered_free_cash_flow_7_yr_cagr"
    dividend_per_share_7_yr_cagr = "dividend_per_share_7_yr_cagr"
    total_revenues_10_yr_cagr = "total_revenues_10_yr_cagr"
    gross_profit_10_yr_cagr = "gross_profit_10_yr_cagr"
    ebitda_10_yr_cagr = "ebitda_10_yr_cagr"
    ebita_10_yr_cagr = "ebita_10_yr_cagr"
    ebit_10_yr_cagr = "ebit_10_yr_cagr"
    earnings_from_cont_ops_10_yr_cagr = "earnings_from_cont_ops_10_yr_cagr"
    net_income_10_yr_cagr = "net_income_10_yr_cagr"
    normalized_net_income_10_yr_cagr = "normalized_net_income_10_yr_cagr"
    diluted_eps_before_extra_10_yr_cagr = "diluted_eps_before_extra_10_yr_cagr"
    accounts_receivable_10_yr_cagr = "accounts_receivable_10_yr_cagr"
    inventory_10_yr_cagr = "inventory_10_yr_cagr"
    net_ppe_10_yr_cagr = "net_ppe_10_yr_cagr"
    common_equity_10_yr_cagr = "common_equity_10_yr_cagr"
    total_assets_10_yr_cagr = "total_assets_10_yr_cagr"
    tangible_book_value_10_yr_cagr = "tangible_book_value_10_yr_cagr"
    cash_from_ops_10_yr_cagr = "cash_from_ops_10_yr_cagr"
    capital_expenditures_10_yr_cagr = "capital_expenditures_10_yr_cagr"
    levered_free_cash_flow_10_yr_cagr = "levered_free_cash_flow_10_yr_cagr"
    unlevered_free_cash_flow_10_yr_cagr = "unlevered_free_cash_flow_10_yr_cagr"
    dividend_per_share_10_yr_cagr = "dividend_per_share_10_yr_cagr"


@unique
class Leverage(Enum):
    """
    财务杠杆
    """
    cash_from_ops_to_curr_liab = "cash_from_ops_to_curr_liab"
    total_debt_to_equity = "total_debt_to_equity"
    total_debt_to_capital = "total_debt_to_capital"
    lt_debt_to_equity = "lt_debt_to_equity"
    lt_debt_to_capital = "lt_debt_to_capital"
    total_liabilities_to_total_assets = "total_liabilities_to_total_assets"
    ebit_to_interest_exp = "ebit_to_interest_exp"
    ebitda_to_interest_exp = "ebitda_to_interest_exp"
    ebitda_capex_to_interest_exp = "ebitda_capex_to_interest_exp"
    total_debt_to_ebitda = "total_debt_to_ebitda"
    net_debt_to_ebitda = "net_debt_to_ebitda"
    total_debt_to_ebitda_capex = "total_debt_to_ebitda_capex"
    net_debt_to_ebitda_capex = "net_debt_to_ebitda_capex"
    capex_as_of_revenues = "capex_as_of_revenues"
    altman_z_score = "altman_z_score"


@unique
class Profitability(Enum):
    """
    盈利能力
    """
    return_on_assets = "return_on_assets"
    return_on_capital = "return_on_capital"
    return_on_equity = "return_on_equity"
    return_on_common_equity = "return_on_common_equity"
    gross_margin = "gross_margin"
    ebitda_margin = "ebitda_margin"
    ebita_margin = "ebita_margin"
    ebit_margin = "ebit_margin"
    sga_margin = "sga_margin"
    earnings_from_cont_ops_margin = "earnings_from_cont_ops_margin"
    net_income_margin = "net_income_margin"
    net_inc_avail_for_common_margin = "net_inc_avail_for_common_margin"
    levered_free_cash_flow_margin = "levered_free_cash_flow_margin"
    unlevered_free_cash_flow_margin = "unlevered_free_cash_flow_margin"
    normalized_net_income_margin = "normalized_net_income_margin"




# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""


class Timeline(object):
    def __init__(self):
        self.latest_time = None
        self.price = None
        self.avg_price = None
        self.volume = None

    def __repr__(self):
        return "Timeline(%s)" % self.__dict__



# -*- coding: utf-8 -*-

import pandas as pd
import six

from tigeropen.common.response import TigerResponse
from tigeropen.common.util.string_utils import get_string

COLUMNS = ['symbol', 'field', 'date', 'value']


class FinancialDailyResponse(TigerResponse):
    def __init__(self):
        super(FinancialDailyResponse, self).__init__()
        self.financial_daily = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(FinancialDailyResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            items = list()
            for item in self.data:
                item_values = dict()
                for key, value in item.items():
                    if isinstance(value, six.string_types):
                        value = get_string(value)
                    item_values[key] = value
                items.append(item_values)
            self.financial_daily = pd.DataFrame(items, columns=COLUMNS)



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""

import six
import pandas as pd
from tigeropen.common.util.string_utils import get_string
from tigeropen.common.response import TigerResponse

COLUMNS = ['symbol', 'ask_price', 'ask_size', 'bid_price', 'bid_size', 'pre_close', 'latest_price', 'latest_time',
           'volume', 'open', 'high', 'low', 'status']
BRIEF_FIELD_MAPPINGS = {'askPrice': 'ask_price', 'askSize': 'ask_size', 'bidPrice': 'bid_price', 'bidSize': 'bid_size',
                        'latestPrice': 'latest_price', 'preClose': 'pre_close', 'latestTime': 'latest_time'}


class StockBriefsResponse(TigerResponse):
    def __init__(self):
        super(StockBriefsResponse, self).__init__()
        self.briefs = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(StockBriefsResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            brief_data = []
            for item in self.data:
                item_values = {}
                for key, value in item.items():
                    if value is None:
                        continue
                    if isinstance(value, six.string_types):
                        value = get_string(value)
                    tag = BRIEF_FIELD_MAPPINGS[key] if key in BRIEF_FIELD_MAPPINGS else key
                    item_values[tag] = value

                brief_data.append([item_values.get(tag) for tag in COLUMNS])

            self.briefs = pd.DataFrame(brief_data, columns=COLUMNS)



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
from __future__ import unicode_literals
import datetime
import uuid

from tigeropen.common.consts import *
from tigeropen.common.consts.params import *
from tigeropen.common.util.common_utils import has_value
from tigeropen.common.util.signature_utils import *
from tigeropen.common.util.web_utils import *
from tigeropen.common.exceptions import *

if not PYTHON_VERSION_3:
    import sys
    reload(sys)
    sys.setdefaultencoding('utf-8')

try:
    from getmac import get_mac_address
except ImportError:
    def get_mac_address():
        return ':'.join(("%012x" % uuid.getnode())[i:i + 2] for i in range(0, 12, 2))


class TigerOpenClient(object):
    """
    client_config：客户端配置，包含tiger_id、应用私钥、老虎公钥等
    logger：日志对象，客户端执行信息会通过此日志对象输出
    """

    def __init__(self, client_config, logger=None):
        self.__config = client_config
        self.__logger = logger
        self.__headers = {
            'Content-type': 'application/json;charset=' + self.__config.charset,
            "Cache-Control": "no-cache",
            "Connection": "Keep-Alive",
            "User-Agent": 'openapi-python-sdk-' + OPEN_API_SDK_VERSION
        }
        self.__device_id = self.__get_device_id()

    """
    内部方法，从params中抽取公共参数
    """

    def __get_common_params(self, params):
        common_params = dict()
        common_params[P_TIMESTAMP] = params[P_TIMESTAMP]
        common_params[P_TIGER_ID] = self.__config.tiger_id
        common_params[P_METHOD] = params[P_METHOD]
        common_params[P_CHARSET] = self.__config.charset
        common_params[P_VERSION] = params[P_VERSION]
        common_params[P_SIGN_TYPE] = self.__config.sign_type
        common_params[P_DEVICE_ID] = self.__device_id
        if has_value(params, P_NOTIFY_URL):
            common_params[P_NOTIFY_URL] = params[P_NOTIFY_URL]
        return common_params

    @staticmethod
    def __get_device_id():
        """
        获取mac地址作为device_id
        :return:
        """
        try:
            return get_mac_address()
        except:
            return None

    """
    内部方法，从params中移除公共参数
    """
    def __remove_common_params(self, params):
        if not params:
            return
        for k in COMMON_PARAM_KEYS:
            if k in params:
                params.pop(k)

    """
    内部方法，通过请求request对象构造请求查询字符串和业务参数
    """
    def __prepare_request(self, request):
        THREAD_LOCAL.logger = self.__logger
        params = request.get_params()
        params[P_TIMESTAMP] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        common_params = self.__get_common_params(params)
        all_params = dict()
        all_params.update(params)
        all_params.update(common_params)
        sign_content = get_sign_content(all_params)
        try:
            sign = sign_with_rsa(self.__config.private_key, sign_content, self.__config.charset)
        except Exception as e:
            raise RequestException('[' + THREAD_LOCAL.uuid + ']request sign failed. ' + str(e))
        all_params[P_SIGN] = sign

        log_url = self.__config.server_url + '?' + sign_content + "&sign=" + sign
        if THREAD_LOCAL.logger:
            THREAD_LOCAL.logger.info('[' + THREAD_LOCAL.uuid + ']request:' + log_url)

        return all_params

    """
    内部方法，解析请求返回结果并做验签
    """

    def __parse_response(self, response_str, timestamp=None):
        if PYTHON_VERSION_3:
            response_str = response_str.decode(self.__config.charset)
        if THREAD_LOCAL.logger:
            THREAD_LOCAL.logger.info('[' + THREAD_LOCAL.uuid + ']response:' + response_str)

        response_content = json.loads(response_str)

        if not self.__config.tiger_public_key or 'sign' not in response_content or not timestamp:
            return response_content

        sign = response_content.get('sign')

        try:
            verify_res = verify_with_rsa(self.__config.tiger_public_key, timestamp.encode('utf-8'),
                                         sign.encode('utf-8'))
        except Exception as e:
            raise ResponseException('[' + THREAD_LOCAL.uuid + ']response sign verify failed. ' + str(e) + ' '
                                    + response_str)
        if not verify_res:
            raise ResponseException('[' + THREAD_LOCAL.uuid + ']response sign verify failed. ' + response_str)

        return response_content

    """
    执行接口请求
    """

    def execute(self, request):
        THREAD_LOCAL.uuid = str(uuid.uuid1())
        query_string = None
        params = self.__prepare_request(request)

        response = do_post(self.__config.server_url, query_string, self.__headers, params, self.__config.timeout,
                           self.__config.charset)

        return self.__parse_response(response, params.get('timestamp'))



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
import logging
import traceback

from tigeropen.trade.domain.order import OrderStatus
from tigeropen.trade.request.model import AccountsParams
from tigeropen.common.response import TigerResponse
from tigeropen.tiger_open_client import TigerOpenClient
from tigeropen.trade.trade_client import TradeClient
from tigeropen.quote.request import OpenApiRequest
from tigeropen.examples.client_config import get_client_config
# from tigeropen.common.consts import Currency, SecurityType
from tigeropen.common.util.contract_utils import stock_contract, option_contract_by_symbol, future_contract, \
     war_contract_by_symbol, iopt_contract_by_symbol
from tigeropen.common.util.order_utils import limit_order, limit_order_with_legs, order_leg, algo_order_params, \
    algo_order

logging.basicConfig(level=logging.INFO,
                    format='%(asctime)s %(levelname)s %(message)s',
                    filemode='a', )
logger = logging.getLogger('TigerOpenApi')

client_config = get_client_config()


def get_account_info():
    from tigeropen.common.consts.service_types import ACCOUNTS
    openapi_client = TigerOpenClient(client_config)
    account = AccountsParams()
    account.account = client_config.account
    request = OpenApiRequest(method=ACCOUNTS, biz_model=account)

    response_content = None
    try:
        response_content = openapi_client.execute(request)
    except Exception as e:
        print(traceback.format_exc())
    if not response_content:
        print("failed to execute")
    else:
        response = TigerResponse()
        response.parse_response_content(response_content)
        if response.is_success():
            print("get response data:" + response.data)
        else:
            print("%d,%s,%s" % (response.code, response.message, response.data))


def get_account_apis():
    openapi_client = TradeClient(client_config, logger=logger)
    openapi_client.get_managed_accounts()
    # 获取订单
    openapi_client.get_orders()
    # 获取未成交订单
    # openapi_client.get_open_orders()
    # 获取已成交订单
    # openapi_client.get_filled_orders(start_time='2019-05-01', end_time='2019-05-21')
    # 获取持仓
    openapi_client.get_positions()
    # 获取资产
    openapi_client.get_assets()


def trade_apis():
    account = client_config.account
    openapi_client = TradeClient(client_config, logger=logger)

    # 通过请求获取合约
    contract = openapi_client.get_contracts('AAPL')[0]
    # contract = openapi_client.get_contract('AAPL', SecurityType.STK, currency=Currency.USD)

    # 本地构造合约
    # stock 股票
    # contract = stock_contract(symbol='AAPL', currency='USD')
    # option 期权
    # contract = option_contract(identifier='AAPL  190118P00160000')
    # contract = option_contract_by_symbol('AAPL', '20200110', strike=280.0, put_call='PUT', currency='USD')
    # future 期货
    # contract = future_contract('CHF', 'USD', '20190617', multiplier=125000, exchange='GLOBEX')
    # war 港股窝轮
    # contract = war_contract_by_symbol('02318', '20200326', 107.08, 'CALL', local_symbol='12616', currency='HKD')
    # iopt 港股牛熊证
    # contract = iopt_contract_by_symbol('02318', '20200420', 87.4, 'CALL', local_symbol='63379', currency='HKD')

    order = openapi_client.create_order(account, contract, 'BUY', 'LMT', 100, limit_price=5.0)
    # 或者本地构造订单对象
    # order = limit_order(account=account, contract=contract, action='BUY', quantity=100, limit_price=5.0)
    openapi_client.place_order(order)
    order_id = order.order_id  # you can operate order via id too

    new_order = openapi_client.get_order(order_id=order.order_id)
    assert order.order_id == new_order.order_id
    openapi_client.modify_order(new_order, quantity=150)
    new_order = openapi_client.get_order(order_id=order_id)
    assert new_order.quantity == 150
    openapi_client.cancel_order(order_id=order_id)
    new_order = openapi_client.get_order(order_id=order_id)
    assert new_order.status == OrderStatus.CANCELLED or new_order.status == OrderStatus.PENDING_CANCEL

    # 预览订单 (下单前后保证金要求, 佣金等预览)
    result = openapi_client.preview_order(order)
    print(result)

    # 限价单 + 附加订单 (仅主订单为限价单时支持附加订单)
    stop_loss_order_leg = order_leg('LOSS', 8.0, time_in_force='GTC')  # 附加止损
    profit_taker_order_leg = order_leg('PROFIT', 12.0, time_in_force='GTC')  # 附加止盈

    main_order = openapi_client.create_order(account, contract, 'BUY', 'LMT', quantity=100, limit_price=10.0,
                                             order_legs=[stop_loss_order_leg, profit_taker_order_leg])
    # 本地构造限价单 + 附加订单
    # main_order = limit_order_with_legs(account, contract, 'BUY', 100, limit_price=10.0,
    # order_legs=[stop_loss_order_leg])

    openapi_client.place_order(main_order)
    print(main_order)
    # 查询主订单所关联的附加订单
    order_legs = openapi_client.get_open_orders(account, parent_id=main_order.order_id)
    print(order_legs)


def algo_order_demo():
    account = client_config.account
    openapi_client = TradeClient(client_config, logger=logger)
    contract = stock_contract(symbol='AAPL', currency='USD')
    params = algo_order_params(start_time='2020-11-19 23:00:00', end_time='2020-11-19 23:50:00', no_take_liq=True,
                               allow_past_end_time=True, participation_rate=0.1)
    order = algo_order(account, contract, 'BUY', 1000, 'VWAP', algo_params=params, limit_price=100.0)
    openapi_client.place_order(order)
    print(order)


if __name__ == '__main__':
    get_account_info()
    get_account_apis()
    trade_apis()



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import six

import pandas as pd
from tigeropen.common.util.string_utils import get_string
from tigeropen.common.response import TigerResponse

COLUMNS = ['symbol', 'time', 'price', 'avg_price', 'pre_close', 'volume', 'trading_session']
TIMELINE_FIELD_MAPPINGS = {'avgPrice': 'avg_price'}


class QuoteTimelineResponse(TigerResponse):
    def __init__(self):
        super(QuoteTimelineResponse, self).__init__()
        self.timelines = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(QuoteTimelineResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            timeline_items = []
            for symbol_item in self.data:
                symbol = symbol_item.get('symbol')
                pre_close = symbol_item.get('preClose')
                if 'preMarket' in symbol_item:  # 盘前
                    pre_markets = symbol_item['preMarket'].get('items')
                    if pre_markets:
                        for item in pre_markets:
                            item_values = self.parse_timeline(item, symbol, pre_close, 'pre_market')
                            timeline_items.append([item_values.get(tag) for tag in COLUMNS])

                if 'intraday' in symbol_item:  # 盘中
                    regulars = symbol_item['intraday'].get('items')
                elif 'items' in symbol_item:
                    regulars = symbol_item['items'][0].get('items')
                else:
                    regulars = None
                if regulars:
                    for item in regulars:
                        item_values = self.parse_timeline(item, symbol, pre_close, 'regular')
                        timeline_items.append([item_values.get(tag) for tag in COLUMNS])

                if 'afterHours' in symbol_item:  # 盘后
                    after_hours = symbol_item['afterHours'].get('items')
                    if after_hours:
                        for item in after_hours:
                            item_values = self.parse_timeline(item, symbol, pre_close, 'after_hours')
                            timeline_items.append([item_values.get(tag) for tag in COLUMNS])

            self.timelines = pd.DataFrame(timeline_items, columns=COLUMNS)

    @staticmethod
    def parse_timeline(item, symbol, pre_close, trading_session):
        item_values = {'symbol': symbol, 'pre_close': pre_close, 'trading_session': trading_session}
        for key, value in item.items():
            if value is None:
                continue
            if isinstance(value, six.string_types):
                value = get_string(value)
            tag = TIMELINE_FIELD_MAPPINGS[key] if key in TIMELINE_FIELD_MAPPINGS else key
            item_values[tag] = value

        return item_values



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import json
import six
from tigeropen.common.response import TigerResponse
from tigeropen.common.util.string_utils import get_string
from tigeropen.common.util.order_utils import get_order_status
from tigeropen.trade.domain.contract import Contract
from tigeropen.trade.domain.order import Order, AlgoParams
from tigeropen.trade.response import CONTRACT_FIELDS

ORDER_FIELD_MAPPINGS = {'parentId': 'parent_id', 'orderId': 'order_id', 'orderType': 'order_type',
                        'limitPrice': 'limit_price', 'auxPrice': 'aux_price', 'avgFillPrice': 'avg_fill_price',
                        'totalQuantity': 'quantity', 'filledQuantity': 'filled', 'lastFillPrice': 'last_fill_price',
                        'realizedPnl': 'realized_pnl', 'secType': 'sec_type',
                        'remark': 'reason',
                        'localSymbol': 'local_symbol', 'originSymbol': 'origin_symbol', 'outsideRth': 'outside_rth',
                        'timeInForce': 'time_in_force', 'openTime': 'order_time', 'latestTime': 'trade_time',
                        'contractId': 'contract_id',
                        'trailStopPrice': 'trail_stop_price', 'trailingPercent': 'trailing_percent',
                        'percentOffset': 'percent_offset', 'identifier': 'identifier', 'algoParameters': 'algo_params'}


class OrdersResponse(TigerResponse):
    def __init__(self):
        super(OrdersResponse, self).__init__()
        self.orders = []
        self._is_success = None

    def parse_response_content(self, response_content, secret_key=None):
        response = super(OrdersResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data:
            data_json = json.loads(self.data)
            if 'items' in data_json:
                for item in data_json['items']:
                    order = OrdersResponse.parse_order(item, secret_key)
                    if order:
                        self.orders.append(order)
            elif 'symbol' in data_json:
                order = OrdersResponse.parse_order(data_json, secret_key)
                if order:
                    self.orders.append(order)

    @staticmethod
    def parse_order(item, secret_key=None):
        contract_fields = {}
        order_fields = {}
        for key, value in item.items():
            if value is None:
                continue
            if isinstance(value, six.string_types):
                value = get_string(value)
            tag = ORDER_FIELD_MAPPINGS[key] if key in ORDER_FIELD_MAPPINGS else key
            if tag in CONTRACT_FIELDS:
                contract_fields[tag] = value
            else:
                order_fields[tag] = value

        contract_id = contract_fields.get('contract_id')
        symbol = contract_fields.get('symbol')
        currency = contract_fields.get('currency')
        sec_type = contract_fields.get('sec_type')
        exchange = contract_fields.get('exchange')
        origin_symbol = contract_fields.get('origin_symbol')
        local_symbol = contract_fields.get('local_symbol')
        expiry = contract_fields.get('expiry')
        strike = contract_fields.get('strike')
        put_call = contract_fields.get('right')
        multiplier = contract_fields.get('multiplier')
        identifier = contract_fields.get('identifier')
        contract = Contract(symbol, currency, contract_id=contract_id, sec_type=sec_type, exchange=exchange,
                            origin_symbol=origin_symbol, local_symbol=local_symbol, expiry=expiry,
                            strike=strike, put_call=put_call, multiplier=multiplier, identifier=identifier)
        account = order_fields.get('account')
        action = order_fields.get('action')
        order_type = order_fields.get('order_type')
        quantity = order_fields.get('quantity')
        limit_price = order_fields.get('limit_price')
        aux_price = order_fields.get('aux_price')
        trail_stop_price = order_fields.get('trail_stop_price')
        trailing_percent = order_fields.get('trailing_percent')
        percent_offset = order_fields.get('percent_offset')
        time_in_force = order_fields.get('time_in_force')
        outside_rth = order_fields.get('outside_rth')
        filled = order_fields.get('filled')
        avg_fill_price = order_fields.get('avg_fill_price')
        commission = order_fields.get('commission')
        realized_pnl = order_fields.get('realized_pnl')
        id = order_fields.get('id')
        order_id = order_fields.get('order_id')
        parent_id = order_fields.get('parent_id')
        status = get_order_status(order_fields.get('status'))
        algo_params = AlgoParams.from_tags(order_fields.get('algo_params'))

        order = Order(account, contract, action, order_type, quantity, limit_price=limit_price, aux_price=aux_price,
                      trail_stop_price=trail_stop_price, trailing_percent=trailing_percent,
                      percent_offset=percent_offset, time_in_force=time_in_force, outside_rth=outside_rth,
                      filled=filled, avg_fill_price=avg_fill_price, commission=commission,
                      realized_pnl=realized_pnl, id=id, order_id=order_id, parent_id=parent_id, algo_params=algo_params)
        if 'order_time' in order_fields:
            order.order_time = order_fields.get('order_time')
        if 'trade_time' in order_fields:
            order.trade_time = order_fields.get('trade_time')
        if 'reason' in order_fields:
            order.reason = order_fields.get('reason')
        if secret_key is not None:
            order.secret_key = secret_key
        order.status = status

        return order




# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""


class AccountProfile(object):
    def __init__(self, account, capability, status):
        self.account = account
        self.capability = capability
        self.status = status

    def __repr__(self):
        """
        String representation for this object.
        """
        return "AccountProfile(%s)" % self.__dict__



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
import json
from tigeropen.common.response import TigerResponse
from tigeropen.trade.domain.account import PortfolioAccount
from tigeropen.common.util.string_utils import camel_to_underline

ACCOUNT_FIELD_MAPPINGS = {'sMA': 'sma', 'updateTime': 'timestamp', 'realizedPnL': 'realized_pnl',
                          'unrealizedPnL': 'unrealized_pnl', 'regTMargin': 'regt_margin', 'regTEquity': 'regt_equity',
                          'cashValue': 'cash', 'initMarginReq': 'initial_margin_requirement',
                          'maintMarginReq': 'maintenance_margin_requirement', 'futuresPnl': 'futures_pnl'}

MARKET_VALUE_FIELD_MAPPINGS = {'updateTime': 'timestamp'}


class AssetsResponse(TigerResponse):
    def __init__(self):
        super(AssetsResponse, self).__init__()
        self.assets = []
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(AssetsResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data:
            data_json = json.loads(self.data)
            if 'items' in data_json:
                for item in data_json['items']:
                    account = item['account']
                    asset = PortfolioAccount(account)
                    summary = asset.summary

                    for key, value in item.items():
                        if value is None:
                            continue
                        tag = ACCOUNT_FIELD_MAPPINGS[key] if key in ACCOUNT_FIELD_MAPPINGS else camel_to_underline(key)
                        if hasattr(summary, tag):
                            setattr(summary, tag, value)
                        elif 'market_values' == tag:
                            if isinstance(value, dict):
                                for sub_key, sub_value in value.items():
                                    currency = sub_key
                                    market_value = asset.market_value(currency=currency)
                                    self._parse_market_value(market_value, sub_value)
                            elif isinstance(value, list):
                                for sub_value in value:
                                    currency = sub_value.get('currency')
                                    if not currency:
                                        continue
                                    market_value = asset.market_value(currency=currency)
                                    self._parse_market_value(market_value, sub_value)
                        elif 'segments' == tag and value:
                            if isinstance(value, dict):
                                for sub_key, sub_value in value.items():
                                    segment_name = sub_key
                                    segment = asset.segment(segment_name=segment_name)
                                    self._parse_segment(segment, sub_value)
                            elif isinstance(value, list):
                                for sub_value in value:
                                    segment_name = sub_value.get('category')
                                    if not segment_name:
                                        continue
                                    segment = asset.segment(segment_name=segment_name)
                                    self._parse_segment(segment, sub_value)

                    self.assets.append(asset)

    @staticmethod
    def _parse_segment(segment, sub_value):
        for segment_key, segment_value in sub_value.items():
            if segment_value is None:
                continue
            if segment_key in ACCOUNT_FIELD_MAPPINGS:
                sub_tag = ACCOUNT_FIELD_MAPPINGS[segment_key]
            else:
                sub_tag = camel_to_underline(segment_key)
            if hasattr(segment, sub_tag):
                setattr(segment, sub_tag, segment_value)

    @staticmethod
    def _parse_market_value(market_value, sub_value):
        for mv_key, mv_value in sub_value.items():
            if mv_value is None:
                continue
            if mv_key in ACCOUNT_FIELD_MAPPINGS:
                sub_tag = ACCOUNT_FIELD_MAPPINGS[mv_key]
            else:
                sub_tag = camel_to_underline(mv_key)
            if hasattr(market_value, sub_tag):
                setattr(market_value, sub_tag, mv_value)



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""
import six
import pandas as pd
from tigeropen.common.response import TigerResponse
from tigeropen.common.util.string_utils import get_string

COLUMNS = ['identifier', 'symbol', 'expiry', 'strike', 'put_call', 'multiplier', 'ask_price', 'ask_size', 'bid_price',
           'bid_size', 'pre_close', 'latest_price', 'volume', 'open_interest']
CHAIN_FIELD_MAPPINGS = {'askPrice': 'ask_price', 'askSize': 'ask_size', 'bidPrice': 'bid_price', 'bidSize': 'bid_size',
                        'latestPrice': 'latest_price', 'openInterest': 'open_interest', 'preClose': 'pre_close',
                        'right': 'put_call'}


class OptionChainsResponse(TigerResponse):
    def __init__(self):
        super(OptionChainsResponse, self).__init__()
        self.chain = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(OptionChainsResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data and isinstance(self.data, list):
            chain_data = []
            for item in self.data:
                symbol = item.get('symbol')
                expiry = item.get('expiry')
                items = item.get('items')
                if symbol and items:
                    for chain_item in items:
                        for call_put_item in chain_item.values():
                            item_values = {'symbol': symbol, 'expiry': expiry}
                            for key, value in call_put_item.items():
                                if value is None:
                                    continue
                                if isinstance(value, six.string_types):
                                    value = get_string(value)
                                if key == 'right':
                                    value = value.upper()
                                tag = CHAIN_FIELD_MAPPINGS[key] if key in CHAIN_FIELD_MAPPINGS else key
                                item_values[tag] = value
                            chain_data.append([item_values.get(tag) for tag in COLUMNS])

            self.chain = pd.DataFrame(chain_data, columns=COLUMNS)



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""
import re

from tigeropen.common.consts import PYTHON_VERSION_3


def add_start_end(key, start_marker, end_marker):
    if key.find(start_marker) < 0:
        key = start_marker + key
    if key.find(end_marker) < 0:
        key = key + end_marker
    return key


def camel_to_underline(hunp_str):
    p = re.compile(r'([a-z]|\d)([A-Z])')
    sub = re.sub(p, r'\1_\2', hunp_str).lower()
    return sub


def get_string(value):
    if PYTHON_VERSION_3:
        return value
    else:
        return value.encode('utf-8')



# -*- coding: utf-8 -*-
"""
Created on 2018/9/20

@author: gaoan
"""

ORDER_NO = "order_no"
PREVIEW_ORDER = "preview_order"
PLACE_ORDER = "place_order"
CANCEL_ORDER = "cancel_order"
MODIFY_ORDER = "modify_order"

"""
账户/资产
"""
ACCOUNTS = "accounts"
ASSETS = "assets"
POSITIONS = "positions"
ORDERS = "orders"
ACTIVE_ORDERS = "active_orders"  # 待成交订单
INACTIVE_ORDERS = "inactive_orders"  # 已撤销订单
FILLED_ORDERS = "filled_orders"  # 已成交订单

"""
合约
"""
CONTRACT = "contract"
CONTRACTS = "contracts"

"""
行情
"""
MARKET_STATE = "market_state"
ALL_SYMBOLS = "all_symbols"
ALL_SYMBOL_NAMES = "all_symbol_names"
BRIEF = "brief"
STOCK_DETAIL = "stock_detail"
TIMELINE = "timeline"
KLINE = "kline"
TRADE_TICK = "trade_tick"
QUOTE_REAL_TIME = "quote_real_time"
QUOTE_SHORTABLE_STOCKS = "quote_shortable_stocks"
QUOTE_STOCK_TRADE = "quote_stock_trade"
QUOTE_DEPTH = "quote_depth"  # level2 深度行情
GRAB_QUOTE_PERMISSION = "grab_quote_permission"  # 抢占行情

# 期权行情
OPTION_EXPIRATION = "option_expiration"
OPTION_CHAIN = "option_chain"
OPTION_BRIEF = "option_brief"
OPTION_KLINE = "option_kline"
OPTION_TRADE_TICK = "option_trade_tick"

# 期货行情
FUTURE_EXCHANGE = "future_exchange"
FUTURE_CONTRACT_BY_CONTRACT_CODE = "future_contract_by_contract_code"
FUTURE_CONTRACT_BY_EXCHANGE_CODE = "future_contract_by_exchange_code"
FUTURE_CONTINUOUS_CONTRACTS = "future_continuous_contracts"
FUTURE_CURRENT_CONTRACT = "future_current_contract"
FUTURE_KLINE = "future_kline"
FUTURE_REAL_TIME_QUOTE = "future_real_time_quote"
FUTURE_TICK = "future_tick"
FUTURE_TRADING_DATE = "future_trading_date"

# 公司行动, 财务数据
FINANCIAL_DAILY = 'financial_daily'
FINANCIAL_REPORT = 'financial_report'
CORPORATE_ACTION = 'corporate_action'

# 行业数据
INDUSTRY_LIST = 'industry_list'
INDUSTRY_STOCKS = 'industry_stocks'
STOCK_INDUSTRY = 'stock_industry'




from .compatibility import *
from .order_methods import *

# 将策略填写在下方
# API 的文档请参考：https://web.itiger.com/tquant/docs/


# 策略初始化方法， 只在开始执行回测时运行一次
def initialize(context):
    context.asset = symbol('AAPL')


def handle_data(context, data):
    print(data.history(context.asset, ['open', 'high', 'low', 'close', 'volume'], 10, '1m'))
    print(data.current(context.asset, ['open', 'high', 'low', 'close', 'volume']))



# -*- coding: utf-8 -*-

import pandas as pd
from tigeropen.common.response import TigerResponse

COLUMNS = ['symbol', 'action_type', 'amount', 'currency', 'announced_date', 'execute_date',
           'record_date', 'pay_date', 'market', 'exchange']
DIVIDEND_FIELD_MAPPINGS = {'actionType': 'action_type', 'announcedDate': 'announced_date',
                           'executeDate': 'execute_date', 'recordDate': 'record_date', 'payDate': 'pay_date',
                           }


class CorporateDividendResponse(TigerResponse):
    def __init__(self):
        super(CorporateDividendResponse, self).__init__()
        self.corporate_dividend = None
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(CorporateDividendResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data:
            items = list()
            for symbol, dividend_items in self.data.items():
                for item in dividend_items:
                    item['symbol'] = symbol
                    items.append(item)
            self.corporate_dividend = pd.DataFrame(items).rename(columns=DIVIDEND_FIELD_MAPPINGS)[COLUMNS]



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""

import json
import six
from tigeropen.common.consts import TradingSession
from tigeropen.common.util.string_utils import get_string
from tigeropen.quote.domain.quote_brief import QuoteBrief, HourTrading
from tigeropen.common.response import TigerResponse

BRIEF_FIELD_MAPPINGS = {'latestPrice': 'latest_price', 'preClose': 'prev_close', 'secType': 'sec_type',
                        'timestamp': 'latest_time', 'askPrice': 'ask_price', 'askSize': 'ask_size',
                        'bidPrice': 'bid_price', 'bidSize': 'bid_size'}


class QuoteBriefResponse(TigerResponse):
    def __init__(self):
        super(QuoteBriefResponse, self).__init__()
        self.briefs = []
        self._is_success = None

    def parse_response_content(self, response_content):
        response = super(QuoteBriefResponse, self).parse_response_content(response_content)
        if 'is_success' in response:
            self._is_success = response['is_success']

        if self.data:
            data_json = json.loads(self.data)
            if 'items' in data_json:
                for item in data_json['items']:
                    brief = QuoteBrief()
                    for key, value in item.items():
                        if value is None:
                            continue
                        if isinstance(value, six.string_types):
                            value = get_string(value)
                        if key == 'hourTrading':
                            hour_trading = HourTrading()
                            for sub_key, sub_value in value.items():
                                if isinstance(sub_value, six.string_types):
                                    sub_value = get_string(sub_value)
                                if sub_key == 'tag':
                                    if sub_value == '盘前':
                                        hour_trading.trading_session = TradingSession.PreMarket
                                    elif sub_value == '盘后':
                                        hour_trading.trading_session = TradingSession.AfterHours
                                else:
                                    sub_tag = BRIEF_FIELD_MAPPINGS[
                                        sub_key] if sub_key in BRIEF_FIELD_MAPPINGS else sub_key
                                    if hasattr(hour_trading, sub_tag):
                                        setattr(hour_trading, sub_tag, sub_value)
                            brief.hour_trading = hour_trading
                        else:
                            tag = BRIEF_FIELD_MAPPINGS[key] if key in BRIEF_FIELD_MAPPINGS else key
                            if hasattr(brief, tag):
                                setattr(brief, tag, value)
                    self.briefs.append(brief)



# -*- coding: utf-8 -*-
"""
Created on 2018/10/31

@author: gaoan
"""


class HourTrading(object):
    def __init__(self):
        self.trading_session = None  # 盘前/盘后
        self.latest_price = None  # 最新价
        self.prev_close = None  # 昨日收盘价
        self.latest_time = None  # 最后交易时间
        self.volume = None  # 成交量

        self.open_price = None  # 开盘价
        self.high_price = None  # 最高价
        self.low_price = None  # 最低价
        self.change = None  # 涨跌额

    def __repr__(self):
        """
        String representation for this object.
        """
        return "HourTrading(%s)" % self.__dict__


class QuoteBrief(object):
    def __init__(self):
        # contract info
        self.symbol = None  # 股票代号
        self.market = None  # 市场代号
        self.name = None  # 股票名称
        self.sec_type = None  # STK 股票, OPT 期权，WAR窝轮，IOPT牛熊证, FUT期货

        self.latest_price = None  # 最新价
        self.prev_close = None  # 昨日收盘价
        self.latest_time = None  # 最后交易时间
        self.volume = None  # 成交量

        self.open_price = None  # 开盘价
        self.high_price = None  # 最高价
        self.low_price = None  # 最低价
        self.change = None  # 涨跌额

        self.bid_price = None  # 卖盘价
        self.bid_size = None  # 卖盘数量
        self.ask_price = None  # 买盘价
        self.ask_size = None  # 买盘数量

        self.halted = None  # 是否停牌 0: 正常,3: 停牌,4: 退市
        self.delay = None  # 延时分钟
        self.auction = None  # 是否竞价时段（仅港股）
        self.expiry = None  # 到期时间（仅港股)

        self.hour_trading = None  # 盘前盘后数据，可能为空（仅美股

    def __repr__(self):
        return "QuoteBrief(%s)" % self.__dict__
